"""Internal module which defines several wrapper classes for Hoc entities.
They are then available as singletons objects in neurodamus.core package
"""

import logging
import os
from contextlib import contextmanager

from .configuration import GlobalConfig, SimConfig


# Singleton, instantiated right below
class _Neuron:
    """A wrapper over the neuron simulator."""

    __name__ = "_Neuron"
    _h = None  # We don't import it at module-level to avoid starting neuron
    _hocs_loaded = set()

    # No new attributes. __setattr__ can rely on it
    __slots__ = ()

    @property
    def h(self):
        """The neuron hoc interpreter, initializing if needed."""
        return self._h or self._init()

    @classmethod
    def _init(cls, mpi=False):
        """Initializes the Neuron simulator."""
        if cls._h is not None:
            return cls._h
        if mpi:
            GlobalConfig.set_mpi()

        from neuron import h, nrn

        cls.__cache = {}
        cls._h = h
        cls.Section = nrn.Section
        cls.Segment = nrn.Segment
        h.load_file("stdrun.hoc")
        h("objref nil")
        h.init()
        return h

    @classmethod
    def load_hoc(cls, mod_name):
        """Loads a hoc module, available in the path."""
        if mod_name in cls._hocs_loaded:
            return
        h = cls._h or cls._init()
        mod_filename = mod_name + ".hoc"
        if not h.load_file(mod_filename):
            msg = (
                f"Can't load HOC library {mod_filename}. "
                f"Consider checking HOC_LIBRARY_PATH: `{os.environ.get('HOC_LIBRARY_PATH')}`"
            )
            raise RuntimeError(msg)
        cls._hocs_loaded.add(mod_name)

    @classmethod
    def require(cls, *hoc_mods):
        """Load a set of hoc mods by name."""
        for mod in hoc_mods:
            cls.load_hoc(mod)
        return cls._h

    @classmethod
    def load_dll(cls, dll_path):
        """Loads a Neuron mod file (typically an .so file in linux)."""
        h = cls._h or cls._init()
        rc = h.nrn_load_dll(dll_path)
        if rc == 0:
            raise RuntimeError(
                f"Cant load MOD dll {dll_path}. Please check LD path and dependencies"
            )

    @contextmanager
    def section_in_stack(self, sec):
        """A contect manager to push and pop a section to the Neuron stack."""
        sec.push()
        yield
        self.h.pop_section()

    # Properties that are not found here are get / set
    # directly in neuron.h
    def __getattr__(self, item):
        # We use a cache since going down to hoc costs at least 10us
        # Cache is not expected to grow very large. Unbounded for the moment
        if item.startswith("__"):
            return object.__getattribute__(self, item)
        self._h or self._init()
        cache = self.__class__.__cache
        obj = cache.get(item)
        if obj is None:
            obj = getattr(self._h, item)
            if type(obj) is not float:
                cache[item] = obj
        return obj

    def __setattr__(self, key, value):
        try:
            object.__setattr__(self, key, value)
        except AttributeError:
            setattr(self.h, key, value)


Neuron = _Neuron()
"""A singleton wrapper for the Neuron library"""


class MComplexLoadBalancer:
    """Wrapper of the load balance Hoc Module with mcomplex."""

    def __init__(self, force_regenerate=False):
        # Can we use an existing mcomplex.dat?
        if force_regenerate or not os.path.isfile("mcomplex.dat"):
            logging.info("Generating mcomplex.dat...")
            self._create_mcomplex()
        else:
            logging.info("Using existing mcomplex.dat")
        self._lb = Neuron.h.LoadBalance()
        self._lb.read_mcomplex()

    @staticmethod
    def _create_mcomplex():
        # Save the dt of the simulation and set the dt for calculating the ExperimentalMechComplex
        # to the default value of 0.025
        prev_dt = Neuron.h.dt
        Neuron.h.dt = SimConfig.default_neuron_dt
        # ExperimentalMechComplex is a complex routine changing many state vars, and cant be reused
        # Therefore here we create a temporary LoadBalance obj
        lb = Neuron.h.LoadBalance()
        lb.ExperimentalMechComplex("StdpWA", "extracel", "HDF5", "Report", "Memory", "ASCII")
        # Revert dt to the old value
        Neuron.h.dt = prev_dt
        # mcomplex changes neuron state and results get different. We re-init
        Neuron.h.init()

    def __getattr__(self, item):
        return getattr(self._lb, item)



import os
from pathlib import Path

import psutil

from neurodamus.utils.pyutils import rmtree

SHM = "/dev/shm"  # noqa: S108


class SHMUtil:
    """Helper class for the SHM file transfer mechanism of CoreNEURON."""

    node_id = -1
    nnodes = -1

    @staticmethod
    def _set_node_info():  # TODO: Replace with MPI SHM communicator
        from neurodamus.core import MPI

        shmdir = SHMUtil.get_datadir_shm("/.__pydamus_nodeinfo_sync")
        path = os.path.join(shmdir, str(MPI.rank))

        # Create SHM folder and files to sync the process count per node
        try:
            os.makedirs(shmdir, exist_ok=True)
            os.close(os.open(path, os.O_CREAT))
        except FileExistsError:
            pass  # Ignore if we have already created the files

        MPI.barrier()

        # Get a filelist sorted by rank ID and store the local node info
        listdir = sorted(os.listdir(shmdir), key=int)
        rank0_node = int(listdir[0])
        nranks_node = len(listdir)

        # Calculate node ID based on the entries that contain a process count
        node_info = MPI.py_gather((nranks_node if MPI.rank == rank0_node else 0), 0)
        if MPI.rank == 0:
            node_info[0] = 0
            for i in range(1, MPI.size):
                node_info[i] = node_info[i - 1] + int(node_info[i] > 0)

        # At this point, the node ID for the rank and # nodes are determined
        SHMUtil.node_id = MPI.py_scatter(node_info, 0)
        SHMUtil.nnodes = MPI.py_broadcast((node_info[-1] + 1) if MPI.rank == 0 else 0, 0)

        rmtree(shmdir)

    @staticmethod
    def _get_approximate_node_rss():
        """Reliable, but imprecise estimate of the nodewide RSS.

        Note: This estimate must work even if MPI ranks can't be associated with
        a physical node.
        """
        vm = psutil.virtual_memory()
        return vm.total - vm.available

    @staticmethod
    def is_node_id_known():
        """Can MPI ranks be associated with physical nodes?"""
        return SHMUtil.get_datadir_shm() is not None

    @staticmethod
    def get_nodewise_rss():
        """For each node the sum of the RSS of all MPI ranks on that node.

        If MPI ranks can't be associated with a node, return `None`.
        """
        from neurodamus.core import MPI, Neuron

        # If we do not have the SHM environment, we can't even know
        # how many nodes there are. Just return `None`.
        if not SHMUtil.is_node_id_known():
            return None

        # Define the node ID for the rank and number of nodes
        if SHMUtil.nnodes < 0:
            SHMUtil._set_node_info()

        # Aggregate the individual memory consumption per node
        process = psutil.Process(os.getpid())
        rss = Neuron.Vector(SHMUtil.nnodes, 0.0)
        rss[SHMUtil.node_id] = process.memory_info().rss

        MPI.allreduce(rss, MPI.SUM)

        return rss.as_numpy().copy()

    @staticmethod
    def get_node_rss():
        # If we do not have the SHM environment, ignore and return an estimate
        if not SHMUtil.is_node_id_known():
            return SHMUtil._get_approximate_node_rss()

        # Return the consumption estimated for the node
        rss = SHMUtil.get_nodewise_rss()
        return rss[SHMUtil.node_id]

    @staticmethod
    def get_mem_avail():
        return psutil.virtual_memory().available

    @staticmethod
    def get_mem_total():
        return psutil.virtual_memory().total

    @staticmethod
    def get_shm_avail():
        return psutil.disk_usage(SHM).free

    @staticmethod
    def get_datadir_shm(datadir=""):
        shmdir = os.environ.get("SHMDIR")
        if not shmdir or not shmdir.startswith(SHM):
            return None

        return str(Path(shmdir) / Path(datadir).resolve().relative_to("/"))

    @staticmethod
    def get_shm_factor():
        factor = os.environ.get("NEURODAMUS_SHM_FACTOR")
        return 0.4 if not factor or not 0.0 <= float(factor) <= 1.0 else float(factor)



"""Compatibility Classes to work similar to HOC types, recreating or wrapping them"""

from array import array


class Vector(array):
    """Behavior similar to Hoc Vector"""

    __slots__ = ()

    def __new__(cls, type_="I", array=None):
        if array is None:
            array = []
        return super().__new__(Vector, type_, array)

    def size(self):
        return len(self)

    @property
    def x(self):
        return self

    def __add__(self, other):
        array.extend(self, other)
        return self


class List(list):  # noqa: FURB189
    """Behavior similar to Hoc List"""

    __slots__ = ()

    def count(self, obj=None):
        return super().count(obj) if obj else len(self)

    def o(self, idx):
        return self[int(idx)]


def hoc_vector(np_array):
    from neuron import h

    hoc_vec = h.Vector(np_array.size)
    hoc_vec.as_numpy()[:] = np_array
    return hoc_vec



from pathlib import Path

from . import NeuronWrapper as Nd
from .configuration import SimConfig


class CompartmentMapping:
    """Interface to register section segment mapping with NEURON."""

    def __init__(self, cell_distributor):
        self.cell_distributor = cell_distributor
        self.pc = Nd.ParallelContext()

    @staticmethod
    def create_section_vectors(section_id, section, secvec, segvec):
        num_segments = 0
        for seg in section:
            secvec.append(section_id)
            segvec.append(seg.node_index())
            num_segments += 1

        return num_segments

    def process_section(self, cell, section, num_electrodes, all_lfp_factors, section_offset):
        secvec, segvec, lfp_factors = Nd.Vector(), Nd.Vector(), Nd.Vector()
        num_segments = 0
        section_attr = getattr(cell._cellref, section[0], None)
        if section_attr:
            for sec in section_attr:
                section_id = cell.get_section_id(sec)
                num_segments += self.create_section_vectors(section_id, sec, secvec, segvec)

        if num_electrodes > 0 and all_lfp_factors.size() > 0 and num_segments > 0:
            start_idx = section_offset * num_electrodes
            end_idx = (section_offset + num_segments) * num_electrodes - 1
            lfp_factors.copy(all_lfp_factors, start_idx, end_idx)

        self.pc.nrnbbcore_register_mapping(
            cell.gid, section[1], secvec, segvec, lfp_factors, num_electrodes
        )
        return num_segments

    def register_mapping(self):
        sections = [
            ("somatic", "soma"),
            ("axonal", "axon"),
            ("basal", "dend"),
            ("apical", "apic"),
            ("AIS", "ais"),
            ("nodal", "node"),
            ("myelinated", "myelin"),
        ]
        gidvec = self.cell_distributor.getGidListForProcessor()
        for activegid in gidvec:
            cell = self.cell_distributor.get_cell(activegid)
            all_lfp_factors = Nd.Vector()
            num_electrodes = 0
            lfp_manager = getattr(self.cell_distributor, "_lfp_manager", None)
            if lfp_manager:
                pop_info = self.cell_distributor.getPopulationInfo(activegid)
                num_electrodes = lfp_manager.get_number_electrodes(activegid, pop_info)
                all_lfp_factors = lfp_manager.read_lfp_factors(activegid, pop_info)

            section_offset = 0
            for section in sections:
                processed_segments = self.process_section(
                    cell, section, num_electrodes, all_lfp_factors, section_offset
                )
                section_offset += processed_segments


class _CoreNEURONConfig:
    """Responsible for managing the configuration of the CoreNEURON simulation.

    It writes the simulation / report configurations and calls the CoreNEURON solver.

    Note: this creates the `CoreConfig` singleton
    """

    default_cell_permute = 0
    artificial_cell_object = None

    @property
    def sim_config_file(self):
        """Get sim config file path to be saved"""
        return str(Path(self.build_path) / "sim.conf")

    @property
    def report_config_file_save(self):
        """Get report config file path to be saved"""
        return str(Path(self.build_path) / "report.conf")

    @property
    def report_config_file_restore(self):
        """Get report config file path to be restored

        We need this file and path for restoring because we cannot recreate it
        from scratch. Only usable when restore exists and is a dir
        """
        return str(Path(SimConfig.restore) / "report.conf")

    @property
    def output_root(self):
        """Get output root from SimConfig"""
        return SimConfig.output_root

    @property
    def datadir(self):
        """Get datadir from SimConfig if not set explicitly"""
        return SimConfig.coreneuron_datadir_path()

    @property
    def build_path(self):
        """Save root folder"""
        return SimConfig.build_path()

    @property
    def restore_path(self):
        """Restore root folder"""
        return SimConfig.restore

    # Instantiates the artificial cell object for CoreNEURON
    # This needs to happen only when CoreNEURON simulation is enabled
    def instantiate_artificial_cell(self):
        self.artificial_cell_object = Nd.CoreNEURONArtificialCell()

    def psolve_core(self, coreneuron_direct_mode=False):
        from neuron import coreneuron

        from . import NeuronWrapper as Nd

        Nd.cvode.cache_efficient(1)
        coreneuron.enable = True
        coreneuron.file_mode = not coreneuron_direct_mode
        coreneuron.sim_config = f"{self.sim_config_file}"
        # set build_path only if the user explicitly asked with --save
        # in this way we do not create 1_2.dat and time.dat if not needed
        if SimConfig.save:
            coreneuron.save_path = self.build_path
        if SimConfig.restore:
            coreneuron.restore_path = self.restore_path

        # Model is already written to disk by calling pc.nrncore_write()
        coreneuron.skip_write_model_to_disk = True
        coreneuron.model_path = f"{self.datadir}"
        Nd.pc.psolve(Nd.tstop)


# Singleton
CoreConfig = _CoreNEURONConfig()



"""Runtime configuration"""

import logging
import os
import os.path
import re
from collections import defaultdict
from enum import Enum
from pathlib import Path

from ._shmutils import SHMUtil
from neurodamus.io.sonata_config import SonataConfig
from neurodamus.utils.logging import log_verbose
from neurodamus.utils.pyutils import ConfigT, StrEnumBase

EXCEPTION_NODE_FILENAME = ".exception_node"
"""A file which controls which rank shows exception"""


class LogLevel:
    ERROR_ONLY = 0
    DEFAULT = 1
    VERBOSE = 2
    DEBUG = 3


class ConfigurationError(Exception):
    """Error due to invalid settings in simulation config
    ConfigurationError should be raised by all ranks to be caught
    properly. Otherwise we might end up with deadlocks.
    For Exceptions that are raised by a single rank Exception
    should be used.
    This is due to the way Exceptions are caught from commands.py.
    """


class GlobalConfig:
    verbosity = LogLevel.DEFAULT
    debug_conn = os.getenv("ND_DEBUG_CONN", "")
    if debug_conn:
        debug_conn = [int(gid) for gid in debug_conn.split(",")]
        verbosity = 3

    @classmethod
    def set_mpi(cls):
        os.environ["NEURON_INIT_MPI"] = "1"


class Feature(Enum):
    Replay = 1
    SpontMinis = 2
    SynConfigure = 3
    Stimulus = 4
    LoadBalance = 5


class CellPermute(StrEnumBase):
    UNPERMUTED = 0  # cpu
    NODE_ADJACENCY = 1  # cpu/gpu

    __mapping__ = [
        ("unpermuted", UNPERMUTED),
        ("node-adjacency", NODE_ADJACENCY),
    ]
    __default__ = UNPERMUTED


class CliOptions(ConfigT):
    cell_permute = None
    report_buffer_size = None
    build_model = None
    simulate_model = True
    output_path = None
    keep_build = False
    lb_mode = None
    modelbuilding_steps = None
    enable_coord_mapping = False
    save = False
    restore = None
    enable_shm = False
    model_stats = False
    simulator = None
    dry_run = False
    num_target_ranks = None
    keep_axon = False
    coreneuron_direct_mode = False
    crash_test = False

    # Restricted Functionality support, mostly for testing

    class NoRestriction:
        """Provide container API, where `in` checks are always True"""

        def __contains__(self, _) -> bool:
            return True

    NoRestriction = NoRestriction()  # Singleton

    restrict_features = NoRestriction  # can also be a list
    restrict_node_populations = NoRestriction
    restrict_connectivity = 0  # no restriction, 1 to disable projections, 2 to disable all


class CircuitConfig(ConfigT):
    name = None
    Engine = None
    nrnPath = ConfigT.REQUIRED  # noqa: N815
    CellLibraryFile = ConfigT.REQUIRED
    METypePath = None
    MorphologyType = None
    MorphologyPath = None
    CircuitTarget = None
    DetailedAxon = False
    PopulationType = None


class RNGConfig(ConfigT):
    global_seed = None
    IonChannelSeed = None
    StimulusSeed = None
    MinisSeed = None
    SynapseSeed = None


class LoadBalanceMode(Enum):
    """An enumeration, inc parser, of the load balance modes."""

    RoundRobin = 0
    WholeCell = 1
    MultiSplit = 2
    Memory = 3

    @classmethod
    def parse(cls, lb_mode):
        """Parses the load balancing mode from a string.
        Options other than WholeCell or MultiSplit are considered RR
        """
        if lb_mode is None:
            return None
        modes = {
            "rr": cls.RoundRobin,
            "roundrobin": cls.RoundRobin,
            "wholecell": cls.WholeCell,
            "loadbalance": cls.MultiSplit,
            "multisplit": cls.MultiSplit,
            "memory": cls.Memory,
        }
        lb_mode_enum = modes.get(lb_mode.lower())
        if lb_mode_enum is None:
            raise ConfigurationError("Unknown load balance mode: " + lb_mode)
        return lb_mode_enum

    class AutoBalanceModeParams:
        """Parameters for auto-selecting a load-balance mode"""

        multisplit_cpu_cell_ratio = 4  # For warning
        cell_count = 1000
        duration = 1000
        mpi_ranks = 200

    @classmethod
    def auto_select(cls, use_neuron, cell_count, duration, auto_params=AutoBalanceModeParams):
        """Simple heuristics for auto selecting load balance"""
        from ._neurodamus import MPI

        lb_mode = LoadBalanceMode.RoundRobin
        reason = ""
        if MPI.size == 1:
            lb_mode = LoadBalanceMode.RoundRobin
            reason = "Single rank - not worth using Load Balance"
        elif use_neuron and MPI.size >= auto_params.multisplit_cpu_cell_ratio * cell_count:
            logging.warning(
                "There's potentially a high number of empty ranks. "
                "To activate multi-split set --lb-mode=MultiSplit"
            )
        elif (
            cell_count > auto_params.cell_count
            and duration > auto_params.duration
            and MPI.size > auto_params.mpi_ranks
        ):
            lb_mode = LoadBalanceMode.WholeCell
            reason = "Simulation size"
        return lb_mode, reason


class _SimConfig:
    """A class initializing several HOC config objects and proxying to simConfig"""

    config_file = None
    cli_options = None
    run_conf = None
    output_root = None
    sonata_circuits = None
    projections = None
    connections = None
    stimuli = None
    reports = None
    modifications = None
    beta_features = None

    # Hoc objects used
    _config_parser = None
    _parsed_run = None
    _simulation_config = None
    _simconf = None
    rng_info = None

    # In principle not all vars need to be required as they'r set by the parameter functions
    simulation_config_dir = None
    current_dir = None
    default_neuron_dt = 0.025
    buffer_time = 25
    save = None
    restore = None
    coreneuron_datadir = None
    extracellular_calcium = None
    use_coreneuron = False
    use_neuron = True
    report_buffer_size = 8  # in MB
    cell_permute = CellPermute.default()
    delete_corenrn_data = False
    modelbuilding_steps = 1
    build_model = True
    simulate_model = True
    loadbal_mode = None
    spike_location = "soma"
    spike_threshold = -30
    dry_run = False
    num_target_ranks = None
    coreneuron_direct_mode = False
    crash_test_mode = False

    _validators = []
    _cell_requirements = {}

    restore_coreneuron = property(lambda self: self.use_coreneuron and bool(self.restore))
    cell_requirements = property(lambda self: self._cell_requirements)

    @classmethod
    def init(cls, config_file, cli_options):
        # Import these objects scope-level to avoid cross module dependency
        from . import NeuronWrapper as Nd

        Nd.init()
        if not os.path.isfile(config_file):
            raise ConfigurationError("Config file not found: " + config_file)
        logging.info("Initializing Simulation Configuration and Validation")

        log_verbose("ConfigFile: %s", config_file)
        log_verbose("CLI Options: %s", cli_options)
        cls.config_file = config_file
        cls._config_parser = cls._init_config_parser(config_file)
        cls._parsed_run = cls._config_parser.parsedRun
        cls._simulation_config = cls._config_parser  # Please refactor me
        cls.simulation_config_dir = os.path.dirname(os.path.abspath(config_file))

        cls.projections = cls._config_parser.parsedProjections
        cls.connections = cls._config_parser.parsedConnects
        cls.stimuli = cls._config_parser.parsedStimuli
        cls.reports = cls._config_parser.parsedReports
        cls.modifications = cls._config_parser.parsedModifications or {}
        cls.beta_features = cls._config_parser.beta_features
        cls.cli_options = CliOptions(**(cli_options or {}))

        cls.dry_run = cls.cli_options.dry_run
        cls.crash_test_mode = cls.cli_options.crash_test
        cls.num_target_ranks = cls.cli_options.num_target_ranks
        # change simulator by request before validator and init hoc config
        if cls.cli_options.simulator:
            cls._parsed_run["Simulator"] = cls.cli_options.simulator

        cls.run_conf = cls._parsed_run
        for validator in cls._validators:
            validator(cls)

        logging.info("Initializing hoc config objects")
        cls._init_hoc_config_objs()

    @classmethod
    def current_dir_path(cls):
        if cls.current_dir:
            return cls.current_dir
        return str(Path.cwd())

    @classmethod
    def build_path(cls):
        """Default to <currend_dir>/build if save is None"""
        return cls.save or str(Path(cls.current_dir_path()) / "build")

    @classmethod
    def coreneuron_datadir_path(cls, create=False):
        """Default to output_root if none is provided"""
        ans = cls.coreneuron_datadir or str(Path(cls.build_path()) / "coreneuron_input")
        if create:
            Path(ans).mkdir(parents=True, exist_ok=True)
        return ans

    @classmethod
    def coreneuron_datadir_restore_path(cls):
        """Default to output_root if none is provided"""
        return str(Path(cls.restore) / "coreneuron_input")

    @classmethod
    def populations_offset_restore_path(cls):
        return str(Path(cls.restore) / "populations_offset.dat")

    @classmethod
    def populations_offset_save_path(cls, create=False):
        """Get polulations_offset path for the save folder.

        Used internally for a save/restore cycle.

        Optional: create pathing folders if necessary
        """
        ans = Path(cls.build_path()) / "populations_offset.dat"
        if create:
            ans.parent.mkdir(parents=True, exist_ok=True)
        return str(ans)

    @classmethod
    def populations_offset_output_path(cls, create=False):
        """Get polulations_offset path for the output folder.

        Used for visualization.

        Optional: create pathing folders if necessary
        """
        ans = Path(cls.output_root) / "populations_offset.dat"
        if create:
            ans.parent.mkdir(parents=True, exist_ok=True)
        return str(ans)

    @classmethod
    def output_root_path(cls, create=False):
        """Get the output_root path

        Create the folder path if required and needed
        """
        outdir = Path(SimConfig.output_root)
        if create:
            outdir.mkdir(parents=True, exist_ok=True)
        return str(outdir)

    @classmethod
    def _init_config_parser(cls, config_file):
        if not config_file.endswith(".json"):
            raise ConfigurationError(
                "Invalid configuration file format. The configuration file must be a .json file."
            )
        try:
            config_parser = SonataConfig(config_file)
        except Exception as e:
            msg = f"Failed to initialize SonataConfig with {config_file}"
            raise ConfigurationError(msg) from e
        return config_parser

    @classmethod
    def _init_hoc_config_objs(cls):
        """Init objects which parse/check configs in the hoc world"""
        from neuron import h

        parsed_run = cls._parsed_run

        cls.rng_info = h.RNGSettings()
        cls.rng_info.interpret(parsed_run, cls.use_coreneuron)
        if "BaseSeed" in parsed_run:
            logging.info("User-defined RNG base seed %s", parsed_run["BaseSeed"])

    @classmethod
    def validator(cls, f):
        """Decorator to register parameters / config validators"""
        cls._validators.append(f)

    @classmethod
    def update_connection_blocks(cls, alias):
        """Convert source destination to real population names

        Args:
            alias: A dict associating alias->real_name's
        """
        from neurodamus.target_manager import TargetSpec  # avoid cyclic deps

        restrict_features = SimConfig.cli_options.restrict_features
        if Feature.SpontMinis not in restrict_features:
            logging.warning("Disabling SpontMinis (restrict_features)")
        if Feature.SynConfigure not in restrict_features:
            logging.warning("Disabling SynConfigure (restrict_features)")

        def update_item(conn, item):
            src_spec = TargetSpec(conn.get(item))
            src_spec.population = alias.get(src_spec.population, src_spec.population)
            conn[item] = str(src_spec)

        new_connections = {}
        for name, conn in cls.connections.items():
            update_item(conn, "Source")
            update_item(conn, "Destination")
            if Feature.SpontMinis not in restrict_features:
                conn.pop("SpontMinis", None)
            if Feature.SynConfigure not in restrict_features:
                conn.pop("SynapseConfigure", None)

            new_connections[name] = conn

        cls.connections = new_connections

    @classmethod
    def check_connections_configure(cls, target_manager):
        check_connections_configure(cls, target_manager)  # top_level

    @classmethod
    def check_cell_requirements(cls, target_manager):
        _input_resistance(cls, target_manager)  # top_level


# Singleton
SimConfig = _SimConfig()


def _check_params(  # noqa: C901
    section_name,
    data,
    required_fields,
    numeric_fields=(),
    non_negatives=(),
    valid_values=None,
    deprecated_values=None,
):
    """Generic function to check a dict-like data set conforms to the field prescription"""
    for param in required_fields:
        if param not in data:
            raise ConfigurationError(
                f"simulation config mandatory param not present: [{section_name}] {param}"
            )
    for param in set(numeric_fields + non_negatives):
        val = data.get(param)
        try:
            val and float(val)
        except ValueError as e:
            raise ConfigurationError(
                f"simulation config param must be numeric: [{section_name}] {param}"
            ) from e
    for param in non_negatives:
        val = data.get(param)
        if val and float(val) < 0:
            raise ConfigurationError(
                f"simulation config param must be positive: [{section_name}] {param}"
            )

    for param, valid in (valid_values or {}).items():
        val = data.get(param)
        if val and val not in valid:
            raise ConfigurationError(
                f"simulation config param value is invalid: [{section_name}] {param} = {val}"
            )

    for param, deprecated in (deprecated_values or {}).items():
        val = data.get(param)
        if val and val in deprecated:
            logging.warning(
                "simulation config param value is deprecated: [%s] %s = %s",
                section_name,
                param,
                val,
            )


@SimConfig.validator
def _run_params(config: _SimConfig):
    required_fields = ("Duration",)
    numeric_fields = ("BaseSeed", "StimulusSeed", "Celsius", "V_Init")
    non_negatives = ("Duration", "Dt", "ModelBuildingSteps")
    _check_params("Run default", config.run_conf, required_fields, numeric_fields, non_negatives)


@SimConfig.validator
def _loadbal_mode(config: _SimConfig):
    cli_args = config.cli_options
    if Feature.LoadBalance not in cli_args.restrict_features:
        logging.warning("Disabled Load Balance (restrict_features)")
        config.loadbal_mode = LoadBalanceMode.RoundRobin
        return
    lb_mode_str = cli_args.lb_mode or config.run_conf.get("RunMode")
    config.loadbal_mode = LoadBalanceMode.parse(lb_mode_str)


@SimConfig.validator
def _projection_params(config: _SimConfig):
    required_fields = ("Path",)
    for name, proj in config.projections.items():
        _check_params("Projection " + name, proj, required_fields, (), ())
        _validate_file_extension(proj.get("Path"))


@SimConfig.validator
def _stimulus_params(config: _SimConfig):
    required_fields = (
        "Mode",
        "Pattern",
        "Duration",
        "Delay",
    )
    numeric_fields = (
        "Dt",
        "RiseTime",
        "DecayTime",
        "AmpMean",
        "AmpVar",
        "MeanPercent",
        "SDPercent",
        "RelativeSkew",
        "AmpStart",
        "AmpEnd",
        "PercentStart",
        "PercentEnd",
        "PercentLess",
        "Mean",
        "Variance",
        "Voltage",
        "RS",
    )
    non_negatives = (
        "Duration",
        "Delay",
        "Rate",
        "Frequency",
        "Width",
        "Lambda",
        "Weight",
        "NumOfSynapses",
        "Seed",
    )
    valid_values = {
        "Mode": ("Current", "Voltage", "Conductance", "spikes"),
        "Pattern": {
            "Hyperpolarizing",
            "Linear",
            "Noise",
            "Pulse",
            "RelativeLinear",
            "RelativeShotNoise",
            "SEClamp",
            "ShotNoise",
            "Sinusoidal",
            "SubThreshold",
            "SynapseReplay",
            "OrnsteinUhlenbeck",
            "NPoisson",
            "NPoissonInhomogeneous",
            "ReplayVoltageTrace",
            "AbsoluteShotNoise",
            "RelativeOrnsteinUhlenbeck",
        },
    }
    deprecated_values = {"Pattern": ("NPoisson", "NPoissonInhomogeneous", "ReplayVoltageTrace")}
    for stim in config.stimuli:
        _check_params(
            "Stimulus " + stim["Name"],
            stim,
            required_fields,
            numeric_fields,
            non_negatives,
            valid_values,
            deprecated_values,
        )


@SimConfig.validator
def _modification_params(config: _SimConfig):
    required_fields = (
        "Target",
        "Type",
    )
    for name, mod_block in config.modifications.items():
        _check_params("Modification " + name, mod_block, required_fields, ())


def make_circuit_config(config_dict, req_morphology=True):
    if config_dict.get("CellLibraryFile") == "<NONE>":
        config_dict["CellLibraryFile"] = False
        config_dict["nrnPath"] = False
        config_dict["MorphologyPath"] = False
    elif config_dict.get("nrnPath") == "<NONE>":
        config_dict["nrnPath"] = False
    _validate_circuit_morphology(config_dict, req_morphology)
    _validate_file_extension(config_dict.get("CellLibraryFile"))
    _validate_file_extension(config_dict.get("nrnPath"))
    return CircuitConfig(config_dict)


def _validate_circuit_morphology(config_dict, required=True):
    morph_path = config_dict.get("MorphologyPath")
    morph_type = config_dict.get("MorphologyType")
    if morph_path is None and required:
        raise ConfigurationError("No morphology path provided (Required!)")
    # Some circuit types may not require morphology files
    if not morph_path:
        log_verbose(" > Morphology src: <Disabled> MorphologyType: %s, ", morph_type or "<None>")
        return
    if morph_type is None:
        logging.warning("MorphologyType not set. Assuming ascii and legacy /ascii subdir")
        morph_type = "asc"
        morph_path = os.path.join(morph_path, "ascii")
        config_dict["MorphologyType"] = morph_type
        config_dict["MorphologyPath"] = morph_path
    assert morph_type in {"asc", "swc", "h5", "hoc"}, "Invalid MorphologyType"
    log_verbose(" > MorphologyType = %s, src: %s", morph_type, morph_path)


def _validate_file_extension(path):
    if not path:
        return
    filepath = path.split(":")[0]  # for sonata, remove the edge_pop name appended to the file path
    if filepath.endswith(".sonata"):
        raise ConfigurationError(
            "*.sonata files are no longer supported, please rename them to *.h5"
        )


@SimConfig.validator
def _circuits(config: _SimConfig):
    from . import EngineBase

    circuit_configs = {}

    for name, circuit_info in config._simulation_config.Circuit.items():
        log_verbose("CIRCUIT %s (%s)", name, circuit_info.get("Engine", "(default)"))
        # Replace name by actual engine class
        circuit_info["Engine"] = EngineBase.get(circuit_info["Engine"])

        circuit_info.setdefault("nrnPath", False)
        if config.cli_options.keep_axon and circuit_info["Engine"].__name__ == "METypeEngine":
            log_verbose("Keeping axons ENABLED")
            circuit_info.setdefault("DetailedAxon", True)

        circuit_configs[name] = make_circuit_config(circuit_info, req_morphology=False)
        circuit_configs[name].name = name

    # Sort so that iteration is deterministic
    config.sonata_circuits = dict(
        sorted(
            circuit_configs.items(),
            key=lambda x: (x[1].Engine.CircuitPrecedence if x[1].Engine else 0, x[0]),
        )
    )


@SimConfig.validator
def _global_parameters(config: _SimConfig):
    from neuron import h

    run_conf = config.run_conf

    config.celsius = run_conf.get("Celsius", 34)
    config.v_init = run_conf.get("V_Init", -65)
    config.extracellular_calcium = run_conf.get("ExtracellularCalcium")
    config.buffer_time = 25 * run_conf.get("FlushBufferScalar", 1)
    config.tstop = run_conf["Duration"]
    h.celsius = config.celsius
    h.set_v_init(config.v_init)
    h.tstop = config.tstop
    config.default_neuron_dt = h.dt
    h.dt = run_conf.get("Dt", h.dt)
    h.steps_per_ms = 1.0 / h.dt
    props = ("celsius", "v_init", "extracellular_calcium", "tstop", "buffer_time")
    log_verbose("Global params: %s", " | ".join(p + f": {getattr(config, p)}" for p in props))
    if "CompartmentsPerSection" in run_conf:
        logging.warning(
            "CompartmentsPerSection is currently not supported. "
            "If needed, please request with a detailed usecase."
        )


@SimConfig.validator
def _set_simulator(config: _SimConfig):
    user_config = config.cli_options
    simulator = config.run_conf.get("Simulator")

    if simulator is None:
        config.run_conf["Simulator"] = simulator = "NEURON"
    if simulator not in {"NEURON", "CORENEURON"}:
        raise ConfigurationError("'Simulator' value must be either NEURON or CORENEURON")
    if simulator == "NEURON" and (
        user_config.build_model is False or user_config.simulate_model is False
    ):
        raise ConfigurationError(
            "Disabling model building or simulation is only compatible with CoreNEURON"
        )

    log_verbose("Simulator = %s", simulator)
    config.use_neuron = simulator == "NEURON"
    config.use_coreneuron = simulator == "CORENEURON"


@SimConfig.validator
def _spike_parameters(config: _SimConfig):
    spike_location = config.run_conf.get("SpikeLocation", "soma")
    if spike_location not in {"soma", "AIS"}:
        raise ConfigurationError("Possible options for SpikeLocation are 'soma' and 'AIS'")
    spike_threshold = config.run_conf.get("SpikeThreshold", -30)
    log_verbose("Spike_Location = %s", spike_location)
    log_verbose("Spike_Threshold = %s", spike_threshold)
    config.spike_location = spike_location
    config.spike_threshold = spike_threshold


_condition_checks = {
    "secondorder": (
        (0, 1, 2),
        ConfigurationError(
            "Time integration method (SecondOrder value) {} is invalid. Valid options are:"
            " '0' (implicitly backward euler),"
            " '1' (Crank-Nicolson) and"
            " '2' (Crank-Nicolson with fixed ion currents)"
        ),
    ),
    "randomize_Gaba_risetime": (
        ("True", "False", "0", "false"),
        ConfigurationError("randomize_Gaba_risetime must be True or False"),
    ),
}


@SimConfig.validator
def _simulator_globals(config: _SimConfig):
    from neuron import h

    # Hackish but some constants only live in the helper
    h.load_file("GABAABHelper.hoc")

    for group in config._simulation_config.Conditions.values():
        for key, value in group.items():
            validator = _condition_checks.get(key)
            if validator:
                config_exception = validator[1] or ConfigurationError(
                    f"Value {value} not valid for key {key}. Allowed: {validator[0]}"
                )
                if value not in validator[0]:
                    raise config_exception

            log_verbose("GLOBAL %s = %s", key, value)

            setattr(h, key, value)

            if "cao_CR" in key and value != config.extracellular_calcium:
                logging.warning(
                    "Value of %s (%s) is not the same as extracellular_calcium (%s)",
                    key,
                    value,
                    config.extracellular_calcium,
                )


@SimConfig.validator
def _second_order(config: _SimConfig):
    second_order = config.run_conf.get("SecondOrder")
    if second_order is None:
        return
    second_order = int(second_order)
    if second_order in {0, 1, 2}:
        from neuron import h

        log_verbose("SecondOrder = %g", second_order)
        config.second_order = second_order
        h.secondorder = second_order
    else:
        raise _condition_checks["secondorder"][1]


@SimConfig.validator
def _single_vesicle(config: _SimConfig):
    if "MinisSingleVesicle" not in config.run_conf:
        return

    from neuron import h

    if not hasattr(h, "minis_single_vesicle_ProbAMPANMDA_EMS"):
        raise NotImplementedError(
            "Synapses don't implement minis_single_vesicle. More recent neurodamus model required."
        )
    minis_single_vesicle = int(config.run_conf["MinisSingleVesicle"])
    log_verbose("minis_single_vesicle = %d", minis_single_vesicle)
    h.minis_single_vesicle_ProbAMPANMDA_EMS = minis_single_vesicle
    h.minis_single_vesicle_ProbGABAAB_EMS = minis_single_vesicle
    h.minis_single_vesicle_GluSynapse = minis_single_vesicle


@SimConfig.validator
def _randomize_gaba_risetime(config: _SimConfig):
    randomize_risetime = config.run_conf.get("RandomizeGabaRiseTime")
    if randomize_risetime is None:
        return
    from neuron import h

    h.load_file("GABAABHelper.hoc")
    if not hasattr(h, "randomize_Gaba_risetime"):
        raise NotImplementedError(
            "Models don't support setting RandomizeGabaRiseTime. "
            "Please load a more recent model or drop the option."
        )
    assert randomize_risetime in {"True", "False", "0", "false"}  # any non-"True" value is negative
    log_verbose("randomize_Gaba_risetime = %s", randomize_risetime)
    h.randomize_Gaba_risetime = randomize_risetime


@SimConfig.validator
def _current_dir(config: _SimConfig):
    curdir = config.run_conf.get("CurrentDir")

    if curdir is None:
        log_verbose("CurrentDir using simulation config path [default]")
        curdir = config.simulation_config_dir
    else:
        if not os.path.isabs(curdir):
            if curdir == ".":
                logging.warning(
                    "Setting CurrentDir to '.' is discouraged and "
                    "shall never be used in production jobs."
                )
            else:
                raise ConfigurationError("CurrentDir: Relative paths not allowed")
            curdir = os.path.abspath(curdir)
        if not os.path.isdir(curdir):
            raise ConfigurationError("CurrentDir doesnt exist: " + curdir)

    log_verbose("CurrentDir = %s", curdir)
    config.run_conf["CurrentDir"] = curdir
    config.current_dir = curdir


@SimConfig.validator
def _output_root(config: _SimConfig):
    """Confirm output_path exists and is usable"""
    output_path = config.run_conf.get("OutputRoot")

    if config.cli_options.output_path not in {None, output_path}:
        output_path = config.cli_options.output_path
    if output_path is None:
        raise ConfigurationError("'OutputRoot' configuration not set")
    if not Path(output_path).is_absolute():
        output_path = Path(config.current_dir_path()) / output_path

    log_verbose("OutputRoot = %s", output_path)
    config.run_conf["OutputRoot"] = str(output_path)
    config.output_root = str(output_path)


@SimConfig.validator
def _check_save(config: _SimConfig):
    cli_args = config.cli_options
    save_path = cli_args.save or config.run_conf.get("Save")

    if not save_path:
        return

    if not config.use_coreneuron:
        raise ConfigurationError("Save-restore only available with CoreNeuron")

    # Handle save
    assert isinstance(save_path, str), "Save must be a string path"
    path_obj = Path(save_path)
    if not path_obj.is_absolute():
        path_obj = Path(config.current_dir) / path_obj

    config.save = str(path_obj)


@SimConfig.validator
def _check_restore(config: _SimConfig):
    restore = config.cli_options.restore or config.run_conf.get("Restore")
    if not restore:
        return

    if not config.use_coreneuron:
        raise ConfigurationError("Save-restore only available with CoreNeuron")

    # sync restore settings to hoc, otherwise we end up with an empty coreneuron_input dir
    config.run_conf["Restore"] = restore

    assert isinstance(restore, str), "Restore must be a string path"
    path_obj = Path(restore)
    if not path_obj.is_absolute():
        path_obj = Path(config.current_dir) / path_obj

    config.restore = str(path_obj)


@SimConfig.validator
def _check_model_build_mode(config: _SimConfig):
    user_config = config.cli_options
    config.build_model = user_config.build_model
    config.simulate_model = user_config.simulate_model

    if config.build_model is True:  # just create the data
        return
    if config.use_coreneuron is False:
        if config.build_model is False:
            raise ConfigurationError("Skipping model build is only available with CoreNeuron")
        # was None
        config.build_model = True
        return

    # CoreNeuron restore is a bit special and already had its input dir checked
    if config.restore:
        config.build_model = False
        return

    # It's a CoreNeuron run. We have to check if build_model is AUTO or OFF
    core_data_location = config.coreneuron_datadir_path()

    try:
        # Ensure that 'sim.conf' and 'files.dat' exist, and that '/dev/shm' was not used
        contents = (Path(core_data_location).parent / "sim.conf").read_text()
        core_data_exists = (
            "datpath='/dev/shm/" not in contents and Path(core_data_location, "files.dat").is_file()
        )
    except FileNotFoundError:
        core_data_exists = False

    if config.build_model in {None, "AUTO"}:
        # If enable-shm option is given we have to rebuild the model and delete any previous files
        # in /dev/shm or gpfs
        # In any case when we start model building any data in the core_data_location_shm if it
        # exists are deleted
        if not core_data_exists:
            core_data_location_shm = SHMUtil.get_datadir_shm(core_data_location)
            data_location = (
                core_data_location_shm
                if (user_config.enable_shm and core_data_location_shm is not None)
                else core_data_location
            )
            logging.info(
                "Valid CoreNeuron input data not found in '%s'. "
                "Neurodamus will proceed to model building.",
                data_location,
            )
            config.build_model = True
        else:
            logging.info(
                "CoreNeuron input data found in %s. Skipping model build", core_data_location
            )
            config.build_model = False

    if not config.build_model and not core_data_exists:
        raise ConfigurationError("Model build DISABLED but no CoreNeuron data found")


@SimConfig.validator
def _keep_coreneuron_data(config: _SimConfig):
    if config.use_coreneuron:
        keep_core_data = False
        if config.cli_options.keep_build or config.run_conf.get("KeepModelData", False) == "True":
            keep_core_data = True
        elif not config.cli_options.simulate_model or config.save:
            logging.warning("Keeping coreneuron data for CoreNeuron following run")
            keep_core_data = True
        config.delete_corenrn_data = not keep_core_data
    log_verbose("delete_corenrn_data = %s", config.delete_corenrn_data)


@SimConfig.validator
def _report_buffer_size(config: _SimConfig):
    user_config = config.cli_options
    if user_config.report_buffer_size is not None:
        report_buffer_size = int(user_config.report_buffer_size)
    else:
        return

    if config.restore is not None:
        raise ConfigurationError("--report-buffer-size cannot be used with --restore")

    assert report_buffer_size > 0, "Report buffer size must be > 0"
    config.report_buffer_size = report_buffer_size


@SimConfig.validator
def _model_building_steps(config: _SimConfig):
    user_config = config.cli_options
    if user_config.modelbuilding_steps is not None:
        ncycles = int(user_config.modelbuilding_steps)
    else:
        return

    assert ncycles > 0, "ModelBuildingSteps set to 0. Required value > 0"

    if not SimConfig.use_coreneuron:
        logging.warning("IGNORING ModelBuildingSteps since simulator is not CORENEURON")
        return

    if "CircuitTarget" not in config.run_conf:
        raise ConfigurationError(
            "Multi-iteration coreneuron data generation requires CircuitTarget"
        )

    logging.info("Splitting Target for multi-iteration CoreNeuron data generation")
    logging.info(" -> Cycles: %d. [src: %s]", ncycles, "CLI")
    config.modelbuilding_steps = ncycles


@SimConfig.validator
def _report_vars(config: _SimConfig):
    """Compartment reports read voltages or i_membrane only. Other types must be summation"""
    mandatory_fields = ("Type", "StartTime", "Target", "Dt", "ReportOn", "Unit", "Format")
    non_negatives = ("StartTime", "EndTime", "Dt")
    report_configs_dict = {}

    for rep_name, rep_config in config.reports.items():
        report_configs_dict[rep_name] = rep_config

        _check_params(
            "Report " + rep_name,
            rep_config,
            mandatory_fields,
            non_negatives=non_negatives,
        )

        if rep_config["Format"] != "SONATA":
            raise ConfigurationError(
                f"Unsupported report format: '{rep_config['Format']}'. Use 'SONATA' instead."
            )

        if (
            config.use_coreneuron
            and rep_config["Type"] == "compartment"
            and rep_config["ReportOn"] not in {"v", "i_membrane"}
        ):
            logging.warning(
                "Compartment reports on vars other than v and i_membrane "
                " are still not fully supported (CoreNeuron)"
            )
    # Overwrite config with a pure dict since we never need underlying hoc map
    config.reports = report_configs_dict


@SimConfig.validator
def _spikes_sort_order(config: _SimConfig):
    order = config.run_conf.get("SpikesSortOrder", "by_time")
    if order not in {"none", "by_time"}:
        raise ConfigurationError(
            f"Unsupported spikes sort order {order}, " + "BBP supports 'none' and 'by_time'"
        )


@SimConfig.validator
def _coreneuron_direct_mode(config: _SimConfig):
    user_config = config.cli_options
    direct_mode = user_config.coreneuron_direct_mode
    if direct_mode:
        if config.use_neuron:
            raise ConfigurationError("--coreneuron-direct-mode is not valid for NEURON")
        if config.modelbuilding_steps > 1:
            logging.warning(
                "--coreneuron-direct-mode not valid for multi-cyle model building, "
                "continue with file mode"
            )
            direct_mode = False
        if config.save or config.restore:
            logging.warning(
                "--coreneuron-direct-mode not valid for save/restore, continue with file mode"
            )
            direct_mode = False

    if direct_mode:
        logging.info("Run CORENEURON direct mode without writing model data to disk")
    config.coreneuron_direct_mode = direct_mode


@SimConfig.validator
def _cell_permute(config: _SimConfig):
    user_config = config.cli_options
    if user_config.cell_permute is not None:
        config.cell_permute = CellPermute.from_string(str(user_config.cell_permute))
    if config.use_neuron and config.cell_permute != CellPermute.UNPERMUTED:
        raise ConfigurationError(
            f"Cell permutation is only available with CoreNEURON. "
            f"--cell-permute={config.cell_permute.to_string()} is invalid with NEURON."
        )


def get_debug_cell_gids(cli_options):
    """Parse the --dump-cell-state option from CLI.

    Supports:
    - A single integer (e.g., "2")
    - A comma-separated list of integers or ranges (e.g., "1,3-5,7")
    - A file path containing one GID per line

    Returns:
        List of 1-based GIDs as integers, or None if not provided.

    Raises:
        ConfigurationError: if the format is invalid or file doesn't exist.
    """
    value = cli_options.get("dump_cell_state") if cli_options else None
    if value is None:
        return None

    def parse_gid_token(token):
        """Parse a single token: an integer or a range (e.g., "3" or "1-4").

        Returns:
            List of integers parsed from the token.

        Raises:
            ConfigurationError: if the token is invalid.
        """
        token = token.strip()
        if re.match(r"^\d+-\d+$", token):
            start, end = map(int, token.split("-"))
            if start > end:
                raise ConfigurationError(f"Invalid range in dump-cell-state: {token}")
            return list(range(start, end + 1))
        if token.isdigit():
            return [int(token)]
        raise ConfigurationError(f"Invalid token in dump-cell-state: {token}")

    try:
        tokens = [str(value)] if isinstance(value, int) else value.split(",")
        gids = []
        for token in tokens:
            gids.extend(parse_gid_token(token))
        gids = [gid + 1 for gid in gids]
        gids = list(dict.fromkeys(gids))  # Remove duplicates while preserving order
    except ValueError as e:
        raise ConfigurationError("Cannot parse dump-cell-state: " + value) from e
    else:
        return gids


def check_connections_configure(config: _SimConfig, target_manager):  # noqa: C901, PLR0912, PLR0915
    """Check connection block configuration and raise warnings for:
    1. Global variable should be set in the Conditions block,
    2. Connection overriding chains (t=0)
    3. Connections with Weight=0 not overridden by delayed (not instantiated)
    4. Partial Connection overriding -> Error
    5. Connections with delay > 0 not overriding anything
    """
    config_assignment = re.compile(r"(\S+)\s*\*?=\s*(\S+)")
    processed_conn_blocks = []
    zero_weight_conns = []
    conn_configure_global_vars = defaultdict(list)

    def get_overlapping_connection_pathway(base_conns, conn):
        for base_conn in reversed(base_conns):
            if target_manager.pathways_overlap(base_conn, conn):
                yield base_conn

    def process_t0_parameter_override(conn):
        if float(conn.get("Weight", 1)) == 0:
            zero_weight_conns.append(conn)
        for overridden_conn in get_overlapping_connection_pathway(processed_conn_blocks, conn):
            conn["_overrides"] = overridden_conn
            if target_manager.pathways_overlap(conn, overridden_conn, equal_only=True):
                overridden_conn["_full_overridden"] = True
            break  # We always compute only against the first overridden block
        processed_conn_blocks.append(conn)

    def process_weight0_override(conn):
        is_overriding = False
        for conn2 in get_overlapping_connection_pathway(zero_weight_conns, conn):
            is_overriding = True
            # If there isn't a full override for zero weights, we must raise exception (later)
            if not conn2.get("_full_overridden"):
                conn2["_full_overridden"] = target_manager.pathways_overlap(
                    conn, conn2, equal_only=True
                )
        if not is_overriding:
            logging.warning(
                "Delayed connection %s is not overriding any weight=0 Connection", conn["_name"]
            )

    def get_syn_config_vars(conn):
        return [var for var, _ in config_assignment.findall(conn.get("SynapseConfigure", ""))]

    def display_overriding_chain(conn):
        logging.warning("Connection %s takes part in overriding chain:", conn["_name"])
        while conn is not None:
            logging.info(
                " -> %-6s %-60.60s Weight: %-8s SpontMinis: %-8s SynConfigure: %s",
                "(base)" if not conn.get("_overrides") else " ^",
                f"{conn['_name']}  {conn['Source']} -> {conn['Destination']}",
                conn.get("Weight", "-"),
                conn.get("SpontMinis", "-"),
                ", ".join(get_syn_config_vars(conn)),
            )
            if conn.get("_visited"):
                break
            conn["_visited"] = True
            conn = conn.get("_overrides")

    logging.info("Checking Connection Configurations")
    all_conn_blocks = config.connections.values()

    # On a first phase process only for t=0
    for name, conn_conf in zip(config.connections, all_conn_blocks):
        conn_conf["_name"] = name
        if float(conn_conf.get("Delay", 0)) > 0.0:
            continue
        for var in get_syn_config_vars(conn_conf):
            if not var.startswith("%s"):
                conn_configure_global_vars[name].append(var)
        # Process all conns to show full override chains to help debug
        process_t0_parameter_override(conn_conf)

    # Second phase: find overridden zero_weights at t > 0
    for conn_conf in all_conn_blocks:
        if float(conn_conf.get("Delay", 0)) > 0:
            process_weight0_override(conn_conf)

    # CHECK 1: Global vars
    if conn_configure_global_vars:
        logging.warning(
            "Global variables in SynapseConfigure. Review the following "
            "connections and move the global vars to Conditions block"
        )
        for name, vars_ in conn_configure_global_vars.items():
            logging.warning(" -> %s: %s", name, vars_)
    else:
        logging.info(" => CHECK No Global vars!")

    # CHECK 2: Block override chains
    if not [
        display_overriding_chain(conn_conf)
        for conn_conf in reversed(processed_conn_blocks)
        if conn_conf.get("_overrides") and not conn_conf.get("_visited")
    ]:
        logging.info(" => CHECK No Block Overrides!")

    # CHECK 3: Weight 0 not/badly overridden
    not_overridden_weight_0 = []
    for conn in zero_weight_conns:
        full_overriden = conn.get("_full_overridden")
        if full_overriden is None:
            not_overridden_weight_0.append(conn)
        elif full_overriden is False:  # incomplete override
            msg = f"Partial Weight=0 override is not supported: Conn {conn['_name']}"
            raise ConfigurationError(msg)
    if not_overridden_weight_0:
        logging.warning(
            "The following connections with Weight=0 are not overridden, "
            "thus won't be instantiated:"
        )
        for conn in not_overridden_weight_0:
            logging.warning(" -> %s", conn["_name"])
    else:
        logging.info(" => CHECK No single Weight=0 blocks!")


def _input_resistance(config: _SimConfig, target_manager):
    prop = "@dynamics:input_resistance"
    for stim in config.stimuli:
        target_or_cs_name = stim["Target"] if "Target" in stim else stim["CompartmentSet"]
        if stim["Mode"] == "Conductance" and stim["Pattern"] in {
            "RelativeShotNoise",
            "RelativeOrnsteinUhlenbeck",
        }:
            # NOTE: use target_manager to read the population names of hoc or NodeSet targets
            population_names = (
                target_manager.get_target(target_or_cs_name).population_names
                if "Target" in stim
                else [target_manager.get_compartment_set(target_or_cs_name).population]
            )
            for population in population_names:
                config._cell_requirements.setdefault(population, set()).add(prop)
                log_verbose(f"[cell] {prop} ({population}:{target_or_cs_name})")



"""neurodamus

The neurodamus package implements the instantiation of simulations in Neuron
based on a configuration file, a.k.a. simulation_config.json
It is deeply based on the HOC implementation, therefore providing python modules like
`node`, `cell_distributor`, etc; and still depends on several low-level HOC files at runtime.
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "devel"

__author__ = "Fernando Pereira <fernando.pereira@epfl.ch>"
__copyright__ = "2018 Blue Brain Project, EPFL"

from .node import Neurodamus, Node

__all__ = ["Neurodamus", "Node"]



"""A wrapper for MorphIO objects. Provides additional neuron
features on top of MorphIO basic morphology handling.
"""

import logging
import os
from dataclasses import dataclass

import numpy as np
from numpy.linalg import eig, norm

"""
    [START] Implementations retrieved from nse/morph-tool (!= hpc/morpho-tool !!!)
    These functions are needed for soma points computation (a la import3d).
    TODO: Once we have something stable, integrate nse/morph-tool
"""
X, Y, R = 0, 1, 3


def split_morphology_path(morphology_path):
    """Split `{collection_path}/{morph_name}.{ext}`"""
    if os.path.exists(morphology_path):
        collection_path = os.path.dirname(morphology_path)
        morph_name, morph_ext = os.path.splitext(os.path.basename(morphology_path))
        return collection_path, morph_name, morph_ext

    collection_path = morphology_path
    while not os.path.exists(collection_path):
        assert collection_path != os.path.dirname(collection_path), "Failed to split path."
        collection_path = os.path.dirname(collection_path)

    morph_name, morph_ext = os.path.splitext(os.path.relpath(morphology_path, collection_path))

    return collection_path, morph_name, morph_ext


def contourcenter(xyz):
    """Python implementation of NEURON code: lib/hoc/import3d/import3d_sec.hoc"""
    POINTS = 101

    points = np.vstack((np.diff(xyz[:, [X, Y]], axis=0), xyz[0, [X, Y]]))
    perim = np.cumsum(np.hstack(((0,), norm(points, axis=1))))[:-1]

    d = np.linspace(0, perim[-1], POINTS)
    new_xyz = np.zeros((POINTS, 3))
    for i in range(3):
        new_xyz[:, i] = np.interp(x=d, xp=perim, fp=xyz[:, i])

    mean = np.mean(new_xyz, axis=0)

    return mean, new_xyz


def get_sides(points, major, minor):
    """Circular permutation of the points so that the point with the largest
    coordinate along the major axis becomes the last point
    tobj = major.c.mul(d.x[i])  ###### uneeded? line 1191
    """
    major_coord, minor_coord = np.dot(points, major), np.dot(points, minor)

    imax = np.argmax(major_coord)
    # pylint: disable=invalid-unary-operand-type
    major_coord, minor_coord = (np.roll(major_coord, -imax), np.roll(minor_coord, -imax))

    imin = np.argmin(major_coord)

    sides = [major_coord[:imin][::-1], major_coord[imin:]]
    rads = [minor_coord[:imin][::-1], minor_coord[imin:]]
    return sides, rads


def make_convex(sides, rads):
    """Keep only points that make path convex"""

    def convex_idx(m):
        """Return index to elements of m that make it convex

        Note: not efficient at the moment
        # now we have the two sides without the min and max points (rads[0]=0)
        # we hope both sides now monotonically increase, i.e. convex
        # make it convex

        """
        idx = np.ones_like(m, dtype=bool)
        last_val = m[-1]
        for i in range(len(m) - 2, -1, -1):
            if m[i] < last_val:
                last_val = m[i]
            else:
                idx[i] = False
        return idx

    for i_side in [0, 1]:
        ci = convex_idx(sides[i_side])
        sides[i_side] = sides[i_side][ci]
        rads[i_side] = rads[i_side][ci]
    return sides, rads


def contour2centroid(mean, points):
    """This follows the function in
         lib/hoc/import3d/import3d_gui.hoc
    most of the comments are from there, so if you want to follow along, it should
    break up the function the same way
    """
    logging.debug("Converting soma contour into a stack of cylinders")

    # find the major axis of the ellipsoid that best fits the shape
    # assuming (falsely in general) that the center is the mean

    points -= mean
    eigen_values, eigen_vectors = eig(np.dot(points.T, points))

    # To be consistent with NEURON eigen vector directions
    eigen_vectors *= -1

    idx = np.argmax(eigen_values)
    major = eigen_vectors[:, idx]
    # minor is normal and in xy plane
    idx = 3 - np.argmin(eigen_values) - np.argmax(eigen_values)
    minor = eigen_vectors[:, idx]
    minor[2] = 0

    sides, rads = get_sides(points, major, minor)
    sides, rads = make_convex(sides, rads)

    tobj = np.sort(np.hstack(sides))
    new_major_coord = np.linspace(tobj[1], tobj[-2], 21)
    rads[0] = np.interp(new_major_coord, sides[0], rads[0])
    rads[1] = np.interp(new_major_coord, sides[1], rads[1])

    points = major * new_major_coord[:, np.newaxis] + mean
    diameters = np.abs(rads[0] - rads[1])

    # avoid 0 diameter ends
    diameters[0] = np.mean(diameters[:2])
    diameters[-1] = np.mean(diameters[-2:])

    return points, diameters


def _to_sphere(neuron):
    """Convert a 3-pts cylinder or a 1-pt sphere into a circular
    contour that represents the same sphere
    """
    radius = neuron.soma.diameters[0] / 2.0
    N = 20
    points = np.zeros((N, 3))
    phase = 2 * np.pi / (N - 1) * np.arange(N)
    points[:, 0] = radius * np.cos(phase)
    points[:, 1] = radius * np.sin(phase)
    points += neuron.soma.points[0]
    neuron.soma.points = points
    neuron.soma.diameters = np.repeat(radius, N)


def single_point_sphere_to_circular_contour(neuron):
    """Transform a single point soma that represents a sphere
    into a circular contour that represents the same sphere
    """
    logging.debug(
        "Converting 1-point soma (sperical soma) to circular contour representing the same sphere"
    )
    _to_sphere(neuron)


"""
    [END] Implementations retrieved from nse/morph-tool (!= hpc/morpho-tool !!!)
"""


@dataclass
class SectionName:
    """A simple container to uniquely identify a NEURON Section by name and ID.

    Attributes:
        name (str): The name of the section (e.g., "soma", "dend", etc.).
                    This corresponds to the section's logical type or label.
        id (int): The index of the section among all sections with the same name.
                  For example, in a list of dendrites, this would identify
                  dend[0], dend[1], etc.

    Example:
        For NEURON's `soma[0]`, the corresponding SectionName would be:

            SectionName(name="soma", id=0)

        This allows unique referencing even in models where multiple sections
        have the same base name.
    """

    name: str
    id: int

    def __str__(self):
        return f"{self.name}[{self.id}]"


class MorphIOWrapper:
    """A class that wraps a MorphIO object and gets everything ready for HOC usage"""

    morph = property(lambda self: self._morph)

    def __init__(self, input_file, options=0):
        self._collection_dir, self._morph_name, self._morph_ext = split_morphology_path(input_file)
        self._options = options
        self._build_morph()
        # This logic is similar to what's in BaseCell, but at this point we are still
        # constructing the cell, so we don't yet have access to a fully initialized instance.
        # Therefore, we cannot reuse the BaseCell implementation directly and need
        # a custom solution here.
        self._section_names = self._get_section_names()
        self._build_sec_typeid_distrib()

    def _build_morph(self):
        """Build immutable morphology, going trough mutable and applying neuron adjustemnts"""
        try:
            # Lazy import morphio since it has an issue with execl
            from morphio import Collection, Morphology, Option, SomaType
        except ImportError as e:
            raise RuntimeError("MorphIO is not available") from e

        collection = Collection(self._collection_dir, extensions=[self._morph_ext])
        options = self._options | Option.nrn_order
        self._morph = collection.load(self._morph_name, options, mutable=True)

        # Re-compute the soma points as they are computed in import3d_gui.hoc
        if self._morph.soma_type not in {SomaType.SOMA_SINGLE_POINT, SomaType.SOMA_SIMPLE_CONTOUR}:
            msg = f"H5 morphology is not supposed to have a soma of type: {self._morph.soma_type}"
            raise Exception(msg)
        logging.debug(
            "(%s, %s, %s) has soma type : %s",
            self._collection_dir,
            self._morph_name,
            self._morph_ext,
            self._morph.soma_type,
        )

        if self._morph.soma_type == SomaType.SOMA_SINGLE_POINT:
            """ See NRN import3d_gui.hoc -> instantiate()
                            -> sphere_rep(xx, yy, zz, dd) """
            single_point_sphere_to_circular_contour(self._morph)
        elif self._morph.soma_type == SomaType.SOMA_SIMPLE_CONTOUR:
            """ See NRN import3d_gui.hoc -> instantiate()
                            -> contour2centroid(xx, yy, zz, dd, sec) """
            mean, new_xyz = contourcenter(self._morph.soma.points)
            self._morph.soma.points, self._morph.soma.diameters = contour2centroid(mean, new_xyz)

        self._morph = Morphology(self._morph)

    def _get_section_names(self) -> list[SectionName]:
        """Returns a list of SectioName

        Relative_index is the index of the section within its type group,
        as expected by the NEURON simulator. NEURON organizes mechanisms
        and pointers (e.g., for synapses or diffusion) into arrays grouped
        by section type (e.g., axon, dendrite), so this relative index
        identifies the section position within its group.
        The list starts with ('soma', 0).
        """
        result = [SectionName("soma", 0)]

        last_type = None
        type_start_index = 0

        for i, sec in enumerate(self._morph.sections, start=1):
            sec_type = self._morph.section_types[sec.id]

            if sec_type != last_type:
                last_type = sec_type
                type_start_index = i

            relative_index = i - type_start_index
            result.append(SectionName(MorphIOWrapper.type2name(sec_type), relative_index))

        return result

    def _build_sec_typeid_distrib(self):
        """Build typeid distribution on top of MorphIO section_types"""
        """
            This will hold np.array that will map the different type ids to the
            start id wrt to section_types and their count. For example, axon type(2)
            starts at section.id 0 and and totals 2724 sections.

                | type_id | start_id  |   count   |
                | ------- | --------- | --------- |
                |      2  |        0  |     2724  |
                |      3  |     2724  |       75  |


            >>> _sec_typeid_distrib
            array([[(2,    0, 2724)],
                   [(3, 2724,   75)]],
                  dtype=[('type_id', '<i8'), ('start_id', '<i8'), ('count', '<i8')])

            Then use it like this:
            >>> _sec_typeid_distrib[['type_id', 'start_id']]
            array([[(2,    0)],
                   [(3, 2724)]],
                  dtype={'names':['type_id','start_id'],....}
        """
        self._sec_typeid_distrib = np.dstack(
            np.unique(self._morph.section_types, return_counts=True, return_index=True)
        )[0]
        self._sec_typeid_distrib = np.concatenate(([(1, -1, 1)], self._sec_typeid_distrib), axis=0)
        self._sec_typeid_distrib.dtype = [("type_id", "<i8"), ("start_id", "<i8"), ("count", "<i8")]

    def morph_as_hoc(self):
        """Uses morphio object to read and generate hoc commands just like import3d_gui.hoc"""
        cmds = []

        """
            We need to get the number of sections for each type in order
            to generate the create commands. E.g.:
                ( soma , 1  )
                ( dend , 52 )
                ( axon , 23 )
                ( apic , 5  )
        """
        # generate create commands
        for [(type_id, count)] in self._sec_typeid_distrib[["type_id", "count"]]:
            tstr = self.type2name(type_id)
            tstr1 = f"create {tstr}[{count}]"
            cmds.append(tstr1)
            tstr1 = self.mksubset(type_id, tstr)
            cmds.append(tstr1)

        cmds.append("forall all.append")

        # generate 3D soma points commands. Order is reversed wrt NEURON's soma points.
        cmds.extend(
            (
                f"soma {{ pt3dadd({p[0]:.8g}, {p[1]:.8g}, {p[2]:.8g}, {d:.8g}) }}"
                for p, d in zip(
                    reversed(self._morph.soma.points), reversed(self._morph.soma.diameters)
                )
            )
        )

        # generate sections connect + their respective 3D points commands
        for i, sec in enumerate(self._morph.sections):
            index = i + 1
            tstr = self._section_names[index]

            if not sec.is_root:
                if sec.parent is not None:
                    parent_index = sec.parent.id + 1
                    tstr1 = self._section_names[parent_index]
                    tstr1 = f"{tstr1} connect {tstr}(0), {1}"
                    cmds.append(tstr1)
            else:
                tstr1 = f"soma connect {tstr}(0), {0.5}"
                cmds.append(tstr1)

                # pt3dstyle does not impact simulation numbers. This will be kept for x-reference.
                # tstr1 = "{} {{ pt3dstyle(1, {:.8g}, {:.8g}, {:.8g}) }}".format
                #                   (tstr, mean[0], mean[1], mean[2])
                # cmds.append(tstr1)

            # 3D point info
            cmds.extend(
                f"{tstr} {{ pt3dadd({p[0]:.8g}, {p[1]:.8g}, {p[2]:.8g}, {d:.8g}) }}"
                for p, d in zip(sec.points, sec.diameters)
            )

        return cmds

    """
         [START] Python versions of import3d_gui.hoc helper functions
         Note: nrn function names will be kept for reference
    """

    _type2name_dict = {1: "soma", 2: "axon", 3: "dend", 4: "apic"}

    @classmethod
    def type2name(cls, type_id):
        """:param type_id: id of section type
        :return: name representation of the section type.
                 If not found in _type2name_dict, then default is:
                    if (type < 0): "minus_{}".format(-type)
                    else: "dend_{}".format(type)
        """
        return cls._type2name_dict.get(type_id) or (
            f"minus_{-type_id}" if type_id < 0 else f"dend_{type_id}"
        )

    _mksubset_dict = {1: "somatic", 2: "axonal", 3: "basal", 4: "apical"}

    @classmethod
    def mksubset(cls, type_id, type_name):
        """:param type_id: id of section type
        :param type_name: the name of the section type
        :return: command to append section type to subset
        """
        tstr = cls._mksubset_dict.get(type_id) or (
            f"minus_{-type_id}set" if type_id < 0 else f"dendritic_{type_id}"
        )

        tstr1 = f'forsec "{type_name}" {tstr}.append'
        return tstr1



import logging

import numpy as np

from .core import NeuronWrapper as Nd
from .core.configuration import ConfigurationError


class LFPManager:
    """Class handling the Online Local Field Potential (LFP) functionality.

    This class is designed to manage the configuration and retrieval of Online
    Local Field Potential (LFP) data used in large-scale neural simulations.
    LFPs represent the aggregate extracellular electrical activity recorded by
    electrodes, reflecting the synchronized activity of nearby neuronal
    populations.

    LFP data in this context is stored in HDF5 files, which must include
    information on:
        - Electrode scaling factors (per compartment contribution to each electrode)
        - Node IDs (cell identifiers contributing to the signal)
        - Offsets (to index subsets of the data per neuron)
    """

    def __init__(self):
        self._lfp_file = None

    def load_lfp_config(self, lfp_weights_file, population_list):
        """Loads lfp weigths from h5 file"""
        logging.info("Reading LFP configuration info from '%s'", lfp_weights_file)
        import h5py

        try:
            self._lfp_file = h5py.File(lfp_weights_file, "r")
        except OSError as e:
            raise ConfigurationError("Error opening LFP electrodes file") from e

        # Check that the file contains the required groups for at least 1 population
        populations_found = []
        for pop_name in population_list:
            req_groups = [
                f"/electrodes/{pop_name}/scaling_factors",
                f"{pop_name}/node_ids",
                f"{pop_name}/offsets",
            ]
            if all(group in self._lfp_file for group in req_groups):
                populations_found.append(pop_name)

        if not populations_found:
            raise ConfigurationError(
                "The LFP weights file does not contain the necessary datasets "
                "'scaling_factors', 'node_ids' and 'offsets' "
                f"in any of the populations {population_list}."
            )

    @staticmethod
    def get_sonata_node_id(gid, population_info):
        return population_info[0], gid - population_info[1] - 1

    def get_node_id_subsets(self, node_id, population_name):
        node_ids = self._lfp_file[population_name]["node_ids"]
        # Look for the index of the node_id
        index = np.where(np.array(node_ids) == node_id)[0][0]
        offsets_dataset = self._lfp_file[population_name]["offsets"]
        electrodes_dataset = self._lfp_file["electrodes"][population_name]["scaling_factors"]
        index_low = offsets_dataset[index]
        index_high = offsets_dataset[index + 1]
        # Get the subset data for the node_id and the section index
        subset_data = electrodes_dataset[index_low:index_high, :]
        return subset_data

    def read_lfp_factors(self, gid, population_info):
        """Reads the local field potential (LFP) factors for a specific gid
        from an HDF5 file and returns the factors as a Nd.Vector.

        Args:
        gid (int): The unique cell identifier
        population_info (Pair(str, int)): Population info ("population_name", population_offset)

        Returns:
        Nd.Vector: A vector containing the LFP factors for the specified gid
        """
        scalar_factors = Nd.Vector()
        if self._lfp_file:
            try:
                population_name, node_id = self.get_sonata_node_id(gid, population_info)
                subset_data = self.get_node_id_subsets(node_id, population_name)
                for electrode_factors in subset_data:
                    scalar_factors.append(Nd.Vector(electrode_factors))
            except (KeyError, IndexError) as e:
                msg = (
                    f"Node id {node_id} missing in the electrodes file"
                    f"for population {population_name}: {e!s}"
                )
                logging.warning(msg)
        return scalar_factors

    def get_number_electrodes(self, gid, population_info):
        """Get number of electrodes of a certain gid"""
        num_electrodes = 0
        if self._lfp_file:
            try:
                population_name, node_id = self.get_sonata_node_id(gid, population_info)
                subset_data = self.get_node_id_subsets(node_id, population_name)
                num_electrodes = subset_data.shape[1]
            except (KeyError, IndexError) as e:
                msg = (
                    f"Node id {node_id} missing in the electrodes file"
                    f"for population {population_name}: {e!s}"
                )
                logging.warning(msg)
        return num_electrodes



import logging
import os
import sys
from time import strftime

from ._engine import EngineBase
from ._mpi import MPI
from ._neuron import _Neuron
from .configuration import GlobalConfig
from neurodamus.utils.logging import log_stage, log_verbose, setup_logging

HOCLIB = "neurodamus"  # neurodamus.hoc should be in HOC_LIBRARY_PATH.
LOG_FILENAME = "pydamus_{}.log".format(strftime("%Y-%m-%d_%Hh%M"))


class _NeuronWrapper(_Neuron):
    """A wrapper class representing an instance of Neuron with the required
    neurodamus hoc and mod modules loaded
    """

    __slots__ = ()
    _pc = None

    @property
    def h(self):
        """The neuron hoc interpreter, initializing if needed"""
        self._pc or self._init()
        return self._h

    @classmethod
    def _init(cls, log_filename=LOG_FILENAME, log_use_color=True):
        if cls._pc is not None:
            return
        # Neurodamus requires MPI. We still respect NEURON_INIT_MPI though
        _Neuron._init(int(os.environ.get("NEURON_INIT_MPI", "1")))  # if needed, sets cls._h

        # Init logging
        log_filename = log_filename or LOG_FILENAME
        if MPI.rank == 0:
            open(log_filename, "w", encoding="utf-8").close()  # Truncate
        MPI.barrier()  # Sync so that all processes see the file
        setup_logging(GlobalConfig.verbosity, log_filename, MPI.rank, use_color=log_use_color)
        log_stage("Initializing Neurodamus... Logfile: " + log_filename)

        # Load mods if not available
        cls._load_nrnmechlibs()
        log_verbose("Mechanisms (mod) library(s) successfully loaded")

        # Load main Hoc
        cls.load_hoc(HOCLIB)

        # Additional libraries introduced in saveUpdate
        sys.path.append(os.environ["HOC_LIBRARY_PATH"])

        # Attempt to instantiate BBSaveState to early detect errors
        cls._h.BBSaveState()
        cls._pc = MPI.pc
        logging.info(" => Neurodamus Mod & Hoc lib loaded.")

        logging.info("Checking for plugins...")
        EngineBase.find_plugins()

    @classmethod
    def _load_nrnmechlibs(cls):
        """Loads the required mods for neurodamus to work

        Two sets are required:
          1. neurodamus mechanisms: "Extensions" for reports, edges, etc
            (built-in or exclusively from NRNMECH_LIB_PATH)
          2. model mechanisms: synapse mechanisms, etc...
            (built-in or coming from BGLIBPY_MOD_LIBRARY_PATH)

        As so, the models may be combined in a library pointed by
        NRNMECH_LIB_PATH which is searched first. Nevertheless
        BGLIBPY_MOD_LIBRARY_PATH must exclusively contain model mechs

        Env vars can also point to several libs separated by ':'
        """

        def check_load_lib(mech, env_lib_path):
            if hasattr(cls._h, mech):
                return True
            mechlib = os.environ.get(env_lib_path)
            if mechlib is None:
                return False

            for libpath in mechlib.split(":"):
                libpath = libpath.strip()
                if os.path.isfile(libpath):
                    logging.info("Loading MECH lib: %s", libpath)
                    cls.load_dll(libpath)
                else:
                    logging.warning("Invalid entry in %s: %s", env_lib_path, libpath)
            return hasattr(cls._h, mech)

        # DEV NOTE: model mods may together with core mods, so check for it first
        # Two independent env vars are required since we want to support "special -python"
        # which might not bring the model (support for split neurodamus) in which case
        # we should load only the model libs pointed by BGLIBPY_MOD_LIBRARY_PATH.

        if not check_load_lib("SonataReport", "NRNMECH_LIB_PATH"):
            logging.error("Could not load neurodamus core mechs from NRNMECH_LIB_PATH")
            sys.exit(1)
        if not check_load_lib("ProbAMPANMDA_EMS", "BGLIBPY_MOD_LIBRARY_PATH"):
            logging.error("Could not load mod library from BGLIBPY_MOD_LIBRARY_PATH")
            sys.exit(1)

    @property
    def pc(self):
        self._pc or self._init()
        return self._pc

    def init(self, **kwargs):
        self._pc or self._init(**kwargs)


# Singleton
NeuronWrapper = _NeuronWrapper()



"""Stimulus implementation where incoming synaptic events are replayed for a single gid"""

from pathlib import Path

import libsonata

from .core.configuration import ConfigurationError
from .utils.multimap import GroupedMultiMap
from .utils.timeit import timeit


class SpikeManager:
    """Holds and manages gid spike time information, specially for Replay.

    A SynapseReplay stim can be used for a single gid that has all the synapses instantiated.
    Given a spikes file from a previous run, this object uses a NetStim object to retrigger
    the synapses at the appropriate time as though the presynaptic cells were present and active.

    Internally the spikes are stored in a :py:class:`neurodamus.utils.multimap.GroupedMultiMap`
    """

    _ascii_spike_dtype = [("time", "double"), ("gid", "uint32")]

    @timeit(name="Replay init")
    def __init__(self, spike_filename, delay=0, population=None):
        """Constructor for SynapseReplay.

        Args:
            spike_filename: path to spike out file.
                if ext is .bin, interpret as binary file; otherwise, interpret as ascii
            delay: delay to apply to spike times
        """
        self._gid_fire_events = None
        # Nd.distributedSpikes = 0  # Wonder the effects of this
        self.open_spike_file(spike_filename, delay, population)

    def open_spike_file(self, filename, delay, population=None):
        """Opens a given spike file.

        Args:
            filename: path to spike out file. Interpret as binary or ascii according to extension
            delay: delay to apply to spike times
        """
        if Path(filename).suffix == (".h5"):
            tvec, gidvec = self._read_spikes_sonata(filename, population)
        else:
            raise ConfigurationError("Spikes input should be a SONATA h5 file")
        if delay:
            tvec += delay

        self._store_events(tvec, gidvec)

    @classmethod
    def _read_spikes_sonata(cls, filename, population):
        spikes_file = libsonata.SpikeReader(filename)
        if population not in spikes_file.get_population_names():
            raise MissingSpikesPopulationError("Spikes population not found: " + population)
        spikes = spikes_file[population]
        spike_dict = spikes.get_dict()
        return spike_dict["timestamps"], spike_dict["node_ids"] + 1

    def _store_events(self, tvec, gidvec):
        """Stores the events in the _gid_fire_events GroupedMultiMap.

        tvec and gidvec arguments should be numpy arrays
        """
        spike_map = GroupedMultiMap(gidvec, tvec)
        if self._gid_fire_events is None:
            self._gid_fire_events = spike_map
        else:
            self._gid_fire_events += spike_map

    def __len__(self):
        return len(self._gid_fire_events)

    def __getitem__(self, gid):
        return self._gid_fire_events.get(gid)

    def __contains__(self, gid):
        return gid in self._gid_fire_events

    def get_map(self):
        """Returns the :py:class:`GroupedMultiMap` with all the spikes."""
        return self._gid_fire_events

    def filter_map(self, pre_gids):
        """Returns a raw dict of pre_gid->spikes for the given pre gids."""
        return {key: self._gid_fire_events[key] for key in pre_gids}


class MissingSpikesPopulationError(Exception):
    """An exception triggered when a given node population is not found, we may want to handle"""



"""Main module for handling and instantiating synaptical connections and gap-junctions"""

from __future__ import annotations

import hashlib
import logging
from collections import defaultdict
from itertools import chain
from os import path as ospath
from typing import TYPE_CHECKING

import numpy as np

from .connection import Connection, ReplayMode
from .core import MPI, NeuronWrapper as Nd, ProgressBarRank0 as ProgressBar, run_only_rank0
from .core.configuration import ConfigurationError, GlobalConfig, SimConfig
from .io.sonata_config import ConnectionTypes
from .io.synapse_reader import SonataReader
from .target_manager import TargetManager, TargetSpec
from .utils import compat
from .utils.logging import VERBOSE_LOGLEVEL, log_all, log_verbose
from .utils.pyutils import bin_search, dict_filter_map, gen_ranges
from .utils.timeit import timeit

if TYPE_CHECKING:
    from .utils.memory import DryRunStats


class ConnectionSet:
    """A dataset of connections.
    Several populations may exist with different seeds
    """

    def __init__(self, src_id, dst_id, conn_factory=Connection):
        # Connections indexed by post-gid, then ordered by pre-gid
        self.src_pop_id = src_id
        self.dst_pop_id = dst_id
        self.src_pop_name = None
        self.dst_pop_name = None
        self.virtual_source = False
        self._conn_factory = conn_factory
        self._connections_map = defaultdict(list)
        self._conn_count = 0

    def __getitem__(self, item):
        return self._connections_map[item]

    def items(self):
        """Iterate over the population as tuples (dst_gid, [connections])"""
        return self._connections_map.items()

    def target_gids(self):
        """Get the list of all targets gids in this Population"""
        return self._connections_map.keys()

    def all_connections(self):
        """Get an iterator over all the connections."""
        return chain.from_iterable(self._connections_map.values())

    def _find_connection(self, sgid, tgid, exact=True):
        """Finds a connection, given its source and destination gids.

        Returns:
            tuple: connection list and index.
                If the element doesnt exist, index depends on 'exact':
                None if exact=True, otherwise the insertion index.
        """
        cell_conns = self._connections_map[tgid]
        pos = 0
        if cell_conns:
            pos = bin_search(cell_conns, sgid, lambda x: x.sgid)
        if exact and (pos == len(cell_conns) or cell_conns[pos].sgid != sgid):
            # Not found
            return cell_conns, None
        return cell_conns, pos

    def get_connection(self, sgid, tgid):
        """Retrieves a connection from the pre and post gids.

        Returns:
            Connection: A connection object if it exists. None otherwise
        """
        conn_lst, idx = self._find_connection(sgid, tgid)
        return None if idx is None else conn_lst[idx]

    def store_connection(self, conn):
        """When we have created a new connection (sgid->tgid), store it
        in order in our structure for faster retrieval later

        Args:
            conn: The connection object to be stored
        """
        cell_conns, pos = self._find_connection(conn.sgid, conn.tgid, exact=False)
        if cell_conns and pos < len(cell_conns) and cell_conns[pos].sgid == conn.sgid:
            logging.error("Attempt to store existing connection: %d->%d", conn.sgid, conn.tgid)
            return
        self._conn_count += 1
        cell_conns.insert(pos, conn)

    def get_or_create_connection(self, sgid, tgid, **kwargs):
        """Returns a connection by pre-post gid, creating if required."""
        conns = self._connections_map[tgid]
        pos = 0
        if conns:
            # optimize for ordered insertion, and handle when sgid is not used
            last_conn = conns[-1]
            if last_conn.sgid in {sgid, None}:
                return last_conn
            if last_conn.sgid < sgid:
                pos = len(conns)
            else:
                pos = bin_search(conns, sgid, lambda x: x.sgid)
                if conns[pos].sgid == sgid:
                    return conns[pos]
        # Not found. Create & insert
        cur_conn = self._conn_factory(sgid, tgid, self.src_pop_id, self.dst_pop_id, **kwargs)
        conns.insert(pos, cur_conn)
        self._conn_count += 1
        return cur_conn

    def get_connections(self, post_gids, pre_gids=None):
        """Get all connections between groups of gids."""
        if isinstance(post_gids, int):
            if pre_gids is None:
                return self._connections_map[post_gids]
            if isinstance(pre_gids, int):
                elem = self.get_connection(pre_gids, post_gids)
                return (elem,) if elem is not None else ()

        post_gid_conn_lists = (
            self._connections_map.values()
            if post_gids is None
            else (self._connections_map[post_gids],)
            if isinstance(post_gids, int)
            else (self._connections_map[tgid] for tgid in post_gids)
        )
        if pre_gids is None:
            return chain.from_iterable(post_gid_conn_lists)
        if isinstance(pre_gids, int):
            # Return a generator which is employing bin search
            return (
                conns[posi]
                for conns in post_gid_conn_lists
                for posi in (bin_search(conns, pre_gids, lambda x: x.sgid),)
                if posi < len(conns) and conns[posi].sgid == pre_gids
            )
        # Generic case. Looks through all conns in selected tgids
        pre_gids = set(pre_gids)
        return (c for conns in post_gid_conn_lists for c in conns if c.sgid in pre_gids)

    def count(self):
        return self._conn_count

    def ids_match(self, population_ids, dst_second=None):
        """Whereas a given population_id selector matches population"""
        if isinstance(population_ids, tuple):
            expr_src, expr_dst = population_ids
        else:
            expr_src, expr_dst = (population_ids, dst_second)
        return (expr_src is None or expr_src == self.src_pop_id) and (
            expr_dst is None or expr_dst == self.dst_pop_id
        )

    def __str__(self):
        return (
            f"<ConnectionSet: {self.src_pop_id}-{self.dst_pop_id} "
            f"({self.src_pop_name}->{self.dst_pop_name})>"
        )

    def __repr__(self):
        return str(self)


class ConnectionManagerBase:
    """An abstract base class common to Synapse and GapJunction connections

    Connection Managers hold and manage connectivity among cell populations.
    For every src-dst pop pairs a new ConnectionManager is created.

    The only case it holds several ConnectionSets is for old-style projections (no population names)
    and possibily future support for multiple edge groups (e.g. from multiple files) so that
    additional connectivity for a pathway can be loaded and configured independently.
    NOTE: self._populations would require a new key format, not (src_pop, dst_pop)
    """

    CONNECTIONS_TYPE = None
    """The type of connections subclasses handle"""

    # Set depending Classes, customizable
    ConnectionSet = ConnectionSet
    SynapseReader = SonataReader
    conn_factory = Connection

    cell_manager = property(lambda self: self._cell_manager)
    src_cell_manager = property(lambda self: self._src_cell_manager)
    is_file_open = property(lambda self: bool(self._synapse_reader))
    src_pop_offset = property(lambda self: self._src_cell_manager.local_nodes.offset)
    target_pop_offset = property(lambda self: self._cell_manager.local_nodes.offset)
    src_node_population = property(lambda self: self._src_cell_manager.population_name)
    dst_node_population = property(lambda self: self._cell_manager.population_name)

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        """Base class c-tor for connections (Synapses & Gap-Junctions)

        Args:
            circuit_conf: A circuit config object where to get the synapse source from
                or None if the ConnecionManager is to be constructed empty
            target_manager: A target manager, where to query target cells
            cell_manager: Manager to query the local target cells
            src_cell_manager: Manager to query the local source cells [default: None -> internal]
            load_offsets: Whether to load synapse offsets

        """
        self._target_manager: TargetManager = target_manager
        self._cell_manager = cell_manager
        self._src_cell_manager = src_cell_manager or cell_manager

        self._populations = {}  # Multiple edge populations support. key is a tuple (src, dst)
        self._cur_population = None

        self._synapse_reader = None
        self._raw_gids = cell_manager.local_nodes.gids(raw_gids=True)
        self._total_connections = 0
        self.circuit_conf = circuit_conf
        self._src_target_filter = None  # filter by src target in all_connect (E.g: GapJ)
        # Load offsets might be required globally. Conn managers shall honor this if possible
        self._load_offsets = kw.get("load_offsets", False)
        # An internal var to enable collection of synapse statistics to a Counter
        self._dry_run_stats: DryRunStats = kw.get("dry_run_stats")
        # For each tgid track "connected" sgids (dry-run)
        self._dry_run_conns = defaultdict(set)

    def __str__(self):
        name = self.__class__.__name__
        return f"<{name:s} | {self._src_cell_manager!s:s} -> {self._cell_manager!s:s}>"

    def open_edge_location(self, syn_source, _circuit_conf, **kw):
        edge_file, *pop = syn_source.split(":")
        pop_name = pop[0] if pop else None
        return self.open_synapse_file(edge_file, pop_name, **kw)

    def open_synapse_file(self, synapse_file, edge_population, *, src_pop_name=None, **_kw):
        """Initializes a reader for Synapses config objects and associated population

        Args:
            synapse_file: The nrn/edge file. For old nrn files it may be a dir.
            edge_population: The population of the edges
            src_name: The source pop name, normally matching that of the source cell manager
        """
        if not ospath.exists(synapse_file):
            raise ConfigurationError(f"Connectivity (Edge) file not found: {synapse_file}")
        if ospath.isdir(synapse_file):
            raise ConfigurationError("Edges source is a directory")

        self._synapse_reader = self._open_synapse_file(synapse_file, edge_population)
        if self._load_offsets and not self._synapse_reader.has_property("synapse_index"):
            raise Exception(
                "Synapse offsets required but not available. "
                "Please use a more recent version of neurodamus-core/synapse-tool"
            )

        self._init_conn_population(src_pop_name)
        self._unlock_all_connections()  # Allow appending synapses from new sources
        return synapse_file

    def _open_synapse_file(self, synapse_file, pop_name):
        logging.debug("Opening Synapse file %s, population: %s", synapse_file, pop_name)
        return self.SynapseReader(
            synapse_file, pop_name, extracellular_calcium=SimConfig.extracellular_calcium
        )

    def _init_conn_population(self, src_pop_name):
        if not src_pop_name:
            src_pop_name = self.src_node_population
        dst_pop_name = self.dst_node_population
        src_pop_id, dst_pop_id = self._compute_pop_ids(src_pop_name, dst_pop_name)

        if self._cur_population and src_pop_id == 0 and not src_pop_name:
            logging.warning(
                "Neither Sonata population nor populationID set. "
                "Edges will be merged with base circuit"
            )

        cur_pop = self.select_connection_set(src_pop_id, dst_pop_id)  # type: ConnectionSet
        cur_pop.src_pop_name = src_pop_name
        cur_pop.dst_pop_name = dst_pop_name
        cur_pop.virtual_source = (
            self._src_cell_manager.is_virtual or src_pop_name != self.src_node_population
        )
        logging.info("Loading connections to population: %s", cur_pop)

    @staticmethod
    def _compute_pop_ids(src_pop_name, dst_pop_name):
        """Compute pop id automatically base on population name."""

        def make_id(node_pop_name):
            pop_hash = hashlib.md5(node_pop_name.encode()).digest()
            return ((pop_hash[1] & 0x0F) << 8) + pop_hash[0]  # id: 12bit hash

        return make_id(src_pop_name), make_id(dst_pop_name)

    def select_connection_set(self, src_pop_id, dst_pop_id):
        """Select the active population of connections given src and dst node pop ids.
        `connect_all()` and `connect_group()` will apply only to the active population.

        Returns: The selected ConnectionSet, eventually created
        """
        self._cur_population = self.get_population(src_pop_id, dst_pop_id)
        return self._cur_population

    def get_population(self, src_pop_id, dst_pop_id):
        """Retrieves a connection set given node src and dst pop ids"""
        pop = self._populations.get((src_pop_id, dst_pop_id))
        if not pop:
            pop = self.ConnectionSet(src_pop_id, dst_pop_id, conn_factory=self.conn_factory)
            self._populations[src_pop_id, dst_pop_id] = pop
        return pop

    # NOTE: Several methods use a selector of the connectivity populations
    # which, to be backwards compat, can be a single ID of the src_population
    # or a tuple to specify source and destination

    def find_populations(self, population_ids):
        """Finds the populations that match a given population selector.

        Args:
            population_ids: A population ids selector. Accepted formats:

                - None: All
                - int: selects matching source population id.
                - tuple(src: Any, dst: Any): Specify source and dest.
                  Each value can also be None, e.g.: (None, 1) selects all
                  populations having post id 1
        """
        if (
            isinstance(population_ids, tuple)
            and population_ids[0] is not None
            and population_ids[1] is not None
        ):
            return [self._populations[population_ids]]
        return [pop for pop in self._populations.values() if pop.ids_match(population_ids)]

    def all_connections(self):
        """Retrieves all the existing connections"""
        return chain.from_iterable(pop.all_connections() for pop in self._populations.values())

    @property
    def connection_count(self):
        return self._total_connections

    @property
    def current_population(self):
        return self._cur_population

    def get_connections(self, post_gids, pre_gids=None, population_ids=None):
        """Retrieves all connections that match post and pre gids eventually
        in a subset of the populations.

        Note: Due to multi-population, a connection may not be unique
        for a given pre-post gid pair. As such get_connection() doesn't
        make sense anymore and this method shall be used instead.

        Args:
            post_gids: The target gids of the connections to search (None = All)
            pre_gids: idem for pre-gids. [Default: all)
            population_ids: A int/tuple of populations ids. Default: all

        """
        for pop in self.find_populations(population_ids):
            yield from pop.get_connections(post_gids, pre_gids)

    def create_connections(self, src_target=None, dst_target=None):
        """Creates connections according to loaded parameters in 'Connection'
        blocks of the config in the currently active ConnectionSet.

        If no Connection block relates to the current population, then load all
        edges. If a single blocks exists with Weight=0, skip creation entirely.

        NOTE: All connections respecting the src_target are retrieved
        and created, even if they use src cells which are NOT instantiated.
        This is to support replay and other stimulus which dont need the src cell.
        If only a subset of connections is wanted, they can be filtered by specifying
        the "source" attribute of the respective connection blocks.

        Args:
            src_target: Target name to restrict creating connections coming from it
            dst_target: Target name to restrict creating connections going into it
        """
        conn_src_spec = TargetSpec(src_target)  # instantiate all from src
        conn_src_spec.population = self.current_population.src_pop_name
        conn_dst_spec = TargetSpec(dst_target or self.cell_manager.circuit_target)
        conn_dst_spec.population = self.current_population.dst_pop_name
        this_pathway = {"Source": str(conn_src_spec), "Destination": str(conn_dst_spec)}
        matching_conns = [
            conn
            for conn in SimConfig.connections.values()
            if self._target_manager.pathways_overlap(conn, this_pathway)
        ]
        if not matching_conns:
            logging.info("No matching Connection blocks. Loading all synapses...")
            self.connect_all()
            return

        # if we have a single connect block with weight=0, skip synapse creation entirely
        if len(matching_conns) == 1 and matching_conns[0].get("Weight") == 0.0:
            logging.warning("SKIPPING Connection create since they have invariably weight=0")
            return

        logging.info("Creating group connections (%d groups match)", len(matching_conns))
        for conn_conf in matching_conns:
            if "Delay" in conn_conf and conn_conf["Delay"] > 0:
                # Delayed connections are for configuration only, not creation
                continue

            # check if we are not supposed to create (only configure later)
            if conn_conf.get("CreateMode") == "NoCreate":
                continue

            conn_src = conn_conf["Source"]
            conn_dst = conn_conf["Destination"]
            synapse_id = conn_conf.get("SynapseID")
            mod_override = conn_conf.get("ModOverride")
            self.connect_group(conn_src, conn_dst, synapse_id, mod_override)

    def configure_connections(self, conn_conf):
        """Configure-only circuit connections according to a config Connection block

        Args:
            conn_conf: The configuration block (dict)
        """
        log_msg = " * Pathway {:s} -> {:s}".format(conn_conf["Source"], conn_conf["Destination"])

        if "Delay" in conn_conf and conn_conf["Delay"] > 0:
            log_msg += f":\t[DELAYED] t={conn_conf['Delay']:g}, weight={conn_conf['Weight']:g}"
            configured_conns = self.setup_delayed_connection(conn_conf)
        else:
            if "SynapseConfigure" in conn_conf:
                log_msg += ":\tconfigure with '{:s}'".format(conn_conf["SynapseConfigure"])
            if "NeuromodStrength" in conn_conf:
                log_msg += "\toverwrite NeuromodStrength = {:g}".format(
                    conn_conf["NeuromodStrength"]
                )
            if "NeuromodDtc" in conn_conf:
                log_msg += "\toverwrite NeuromodDtc = {:g}".format(conn_conf["NeuromodDtc"])
            configured_conns = self.configure_group(conn_conf)

        all_ranks_total = MPI.allreduce(configured_conns, MPI.SUM)
        if all_ranks_total > 0:
            logging.info(log_msg)
            logging.info(" => Configured %s connections", all_ranks_total)

    def setup_delayed_connection(self, conn_config):
        raise NotImplementedError(
            f"Manager {self.__class__.__name__} doesn't implement delayed connections"
        )

    def connect_all(self, weight_factor=1):
        """For every gid access its synapse parameters and instantiate all synapses.

        Args:
            weight_factor: Factor to scale all netcon weights (default: 1)
            only_gids: Create connections only for these tgids (default: Off)
        """
        if SimConfig.dry_run:
            syn_count = self._get_conn_stats(None)
            log_all(VERBOSE_LOGLEVEL, "[Rank %d] Synapse count: %d", MPI.rank, syn_count)
            self._dry_run_stats.synapse_counts[self.CONNECTIONS_TYPE] += syn_count
            return

        conn_options = {"weight_factor": weight_factor}
        pop = self._cur_population

        for sgid, tgid, syns_params, extra_params, offset in self._iterate_conn_params(
            src_target=self._src_target_filter, dst_target=None, show_progress=True
        ):
            if self._load_offsets:
                conn_options["synapses_offset"] = extra_params["synapse_index"][0]
            # Create all synapses. No need to lock since the whole file is consumed
            cur_conn = pop.get_or_create_connection(sgid, tgid, **conn_options)
            self._add_synapses(cur_conn, syns_params, None, offset)

    def connect_group(
        self, conn_source, conn_destination, synapse_type_restrict=None, mod_override=None
    ):
        """Instantiates pathway connections & synapses given src-dst

        Args:
            conn_source (str): The target name of the source cells
            conn_destination (str): The target of the destination cells
            synapse_type_restrict(int): Create only given synType synapses
            mod_override (str): ModOverride given for this connection group
        """
        conn_kwargs = {}
        conn_pop = self._cur_population
        dst_pop_name = self.dst_node_population
        src_pop_name = self.src_node_population
        logging.debug("Connecting group %s -> %s", conn_source, conn_destination)
        src_tspec = TargetSpec(conn_source)
        dst_tspec = TargetSpec(conn_destination)
        src_target = src_tspec.name and self._target_manager.get_target(src_tspec, src_pop_name)
        dst_target = dst_tspec.name and self._target_manager.get_target(dst_tspec, dst_pop_name)

        if (src_target and src_target.is_void()) or (dst_target and dst_target.is_void()):
            logging.debug(
                "Skip void connectivity for current connectivity: %s - %s",
                conn_source,
                conn_destination,
            )
            return

        if SimConfig.dry_run:
            syn_count = self._get_conn_stats(dst_target, src_target)
            log_all(VERBOSE_LOGLEVEL, "%s-> %s: %d", conn_pop.src_name, conn_destination, syn_count)
            self._dry_run_stats.synapse_counts[self.CONNECTIONS_TYPE] += syn_count
            return

        for sgid, tgid, syns_params, extra_params, offset in self._iterate_conn_params(
            src_target, dst_target, show_progress=None, mod_override=mod_override
        ):
            if sgid == tgid:
                logging.warning("Making connection within same Gid: %d", sgid)
            if self._load_offsets:
                conn_kwargs["synapses_offset"] = extra_params["synapse_index"][0]

            cur_conn = conn_pop.get_or_create_connection(sgid, tgid, **conn_kwargs)
            if cur_conn.locked:
                continue
            self._add_synapses(cur_conn, syns_params, synapse_type_restrict, offset)
            cur_conn.locked = True

    def _add_synapses(self, cur_conn: Connection, syns_params, syn_type_restrict=None, base_id=0):
        if syn_type_restrict:
            syns_params = syns_params[syns_params["synType"] != syn_type_restrict]
        if len(syns_params) == 0:
            return
        if SimConfig.crash_test_mode:
            cur_conn.add_single(self._cell_manager, syns_params[0], base_id)
        else:
            cur_conn.add_synapses(self._target_manager, syns_params, base_id)

    class ConnDebugger:
        __slots__ = ["yielded_src_gids"]

        def __init__(self):
            self.yielded_src_gids = compat.array("i") if GlobalConfig.debug_conn else None

        def register(self, sgid, base_tgid, syns_params):
            if not (debug_conn := GlobalConfig.debug_conn):
                return
            if debug_conn == [base_tgid]:
                self.yielded_src_gids.append(sgid)
            elif debug_conn == [sgid, base_tgid]:
                log_all(
                    logging.DEBUG, "Connection (%d-%d). Params:\n%s", sgid, base_tgid, syns_params
                )

        def __del__(self):
            if self.yielded_src_gids:
                log_all(logging.DEBUG, "Source GIDs for debug cell: %s", self.yielded_src_gids)

    @staticmethod
    def _get_allowed_ranges(src_target, sgids, sgids_ranges, conn_count):
        """Return n_yielded_conns and allowed_ranges, handling src_target=None.

        Helper function for _iterate_conn_params
        """
        if src_target:
            unique_sgids = sgids[sgids_ranges[:-1]]
            allowed_sgids = set(unique_sgids[src_target.contains(unique_sgids, raw_gids=True)])
            allowed_ranges = [
                (sgids_ranges[i], sgids_ranges[i + 1])
                for i in range(conn_count)
                if sgids[sgids_ranges[i]] in allowed_sgids
            ]
            n_yielded_conns = len(allowed_sgids)
        else:
            n_yielded_conns = conn_count
            allowed_ranges = [(sgids_ranges[i], sgids_ranges[i + 1]) for i in range(conn_count)]

        return n_yielded_conns, allowed_ranges

    @staticmethod
    def _compute_sgids_ranges(syns_params):
        """Compute source GIDs, their change points, and total connections count.

        Helper function for _iterate_conn_params
        """
        sgids = syns_params[syns_params.dtype.names[0]].astype("int64")
        sgids_ranges = np.diff(sgids, prepend=np.nan, append=np.nan).nonzero()[0]
        conn_count = len(sgids_ranges) - 1
        return sgids, sgids_ranges, conn_count

    def _get_extra_fields(self, base_tgid):
        """Get extra fields for the synapse parameters, e.g. synapse_index.

        Helper function for _iterate_conn_params
        """
        if self._load_offsets:
            syn_index = self._synapse_reader.get_property(base_tgid, "synapse_index")
            return {"synapse_index": syn_index}
        return {}

    def _iterate_conn_params(  # noqa: PLR0914
        self,
        src_target,
        dst_target,
        show_progress=None,
        mod_override=None,
    ):
        """A generator which loads synapse data and yields tuples(sgid, tgid, synapses)

        Args:
            src_target: the target to filter the source cells, or None
            dst_target: the target to filter the destination cells, or None
            show_progress: Display a progress bar as tgids are processed
        """
        AUTO_PROGRESS_THRESHOLD = 50
        if (src_target and src_target.is_void()) or (dst_target and dst_target.is_void()):
            return

        gids = self._raw_gids

        if dst_target:
            gids = np.intersect1d(gids, dst_target.gids(raw_gids=True))

        created_conns_0 = self._cur_population.count()
        sgid_offset = self.src_pop_offset
        tgid_offset = self.target_pop_offset

        self._synapse_reader.configure_override(mod_override)
        self._synapse_reader.preload_data(gids, minimal_mode=SimConfig.cli_options.crash_test)

        # NOTE: This routine is quite critical, sitting at the core of synapse processing
        # so it has been carefully optimized with numpy vectorized operations, even if
        # it might lose some readability.
        # For each tgid we obtain the synapse parameters as a record array. We then split it,
        # without copying, yielding ranges (views) of it.
        if show_progress is None:
            show_progress = len(gids) >= AUTO_PROGRESS_THRESHOLD

        gids = ProgressBar.iter(gids, name="Loading") if show_progress else gids

        for base_tgid in gids:
            tgid = base_tgid + tgid_offset
            syns_params = self._synapse_reader.get_synapse_parameters(base_tgid)
            logging.debug("GID %d Syn count: %d", tgid, len(syns_params))

            sgids, sgids_ranges, conn_count = self._compute_sgids_ranges(syns_params)
            conn_debugger = self.ConnDebugger()
            if conn_count == 0:
                logging.debug("No synapses for GID %d. Nothing to do.", tgid)
                continue

            extra_fields = self._get_extra_fields(base_tgid)

            # We yield ranges of contiguous parameters belonging to the same connection,
            # and given we have data for a single tgid, enough to group by sgid.
            # The first row of a range is found by numpy.diff

            n_yielded_conns, allowed_ranges = self._get_allowed_ranges(
                src_target, sgids, sgids_ranges, conn_count
            )

            for range_start, range_end in allowed_ranges:
                sgid = int(sgids[range_start])
                final_sgid = sgid + sgid_offset
                syn_params = syns_params[range_start:range_end]
                extra_params = (
                    extra_fields
                    and {  # reuse empty {}. Dont modify later!
                        name: prop[range_start:range_end] for name, prop in extra_fields.items()
                    }
                )
                conn_debugger.register(sgid, base_tgid, syn_params)
                yield final_sgid, tgid, syn_params, extra_params, range_start

            logging.debug(
                " > Yielded %d out of %d connections. (Filter by src Target: %s)",
                n_yielded_conns,
                conn_count,
                src_target and src_target.name,
            )

        created_conns = self._cur_population.count() - created_conns_0
        self._total_connections += created_conns

        all_created = MPI.allreduce(created_conns, MPI.SUM)
        if all_created:
            pathway_repr = "[ALL]"
            if src_target and dst_target:
                pathway_repr = f"Pathway {src_target.name} -> {dst_target.name}"
            logging.info(" * %s. Created %d connections", pathway_repr, all_created)

    def _get_conn_stats(self, dst_target, src_target=None):  # noqa: PLR0914
        """Estimates the number of synapses for the given destination and source nodesets

        Args:
            dst_nodeset: The target to estimate synapses for
            src_nodeset: The source nodes allowed for the given synapses

        Returns:
            The estimated number of synapses which would be created

        """
        BLOCK_BASE_SIZE = 5000
        SAMPLED_CELLS_PER_BLOCK = 100

        # Get the raw gids for the destination target (in this rank)
        local_gids = dst_target.get_local_gids(raw_gids=True) if dst_target else self._raw_gids
        if not len(local_gids):  # Target is empty in this rank
            logging.debug("Skipping group: no cells!")
            return 0

        total_estimate = 0
        gids_per_metype = self._dry_run_stats.pop_metype_gids[self.dst_node_population]

        # NOTE:
        #  - Estimation (and extrapolation) is performed per metype since properties can vary
        #  - Consider only the cells for the current target

        for metype, all_me_gids in gids_per_metype.items():
            me_gids = set(all_me_gids).intersection(local_gids)
            me_gids_count = len(me_gids)
            if not me_gids_count:
                logging.debug("Skipping metype '%s': no cells!", metype)
                continue

            logging.debug("Metype %s", metype)
            me_gids = np.fromiter(me_gids, dtype="uint32")

            # NOTE:
            # Process the first 100 cells from increasingly large blocks
            #  - Takes advantage of data locality
            #  - Blocks increase as a geometric progression for handling very large sets

            metype_estimate = 0

            for (
                start,
                stop,
            ) in gen_ranges(me_gids_count, BLOCK_BASE_SIZE, block_increase_rate=1.1):
                logging.debug(" - Processing range %d:%d", start, stop)
                block_len = stop - start
                sample = me_gids[start : (start + SAMPLED_CELLS_PER_BLOCK)]
                sample_len = len(sample)
                if not sample_len:
                    continue

                sample_counts = self._synapse_reader.get_conn_counts(sample)
                total_connections = 0
                selected_conn_count = 0
                new_conn_count = 0  # Let's count those which were not "created" before
                new_syns_count = 0

                for tgid, tgid_conn_counts in sample_counts.items():
                    total_connections += len(tgid_conn_counts)
                    if src_target:
                        conn_sgids = np.fromiter(tgid_conn_counts.keys(), dtype="uint32")
                        sgids_in_target = conn_sgids[src_target.contains(conn_sgids, raw_gids=True)]
                    else:
                        sgids_in_target = tgid_conn_counts.keys()

                    selected_conn_count += len(sgids_in_target)
                    tgid_connected_sgids = self._dry_run_conns[tgid]

                    for sgid in sgids_in_target:
                        if sgid not in tgid_connected_sgids:
                            new_conn_count += 1
                            new_syns_count += tgid_conn_counts[sgid]
                            tgid_connected_sgids.add(int(sgid))

                logging.debug(
                    " - Connections (new/selected/total): %d / %d / %d ",
                    new_conn_count,
                    selected_conn_count,
                    total_connections,
                )
                block_syns_per_cell = new_syns_count / sample_len
                logging.debug(" - Synapses: %d (Avg: %.2f)", new_syns_count, block_syns_per_cell)
                metype_estimate += block_syns_per_cell * block_len

            # Info on the whole metype (subject to selected target)
            # Due to the fact that the same metype might be target of several projections
            #   we have to sum the averages
            average_syns_per_cell = metype_estimate / me_gids_count
            self._dry_run_stats.metype_cell_syn_average[metype] += average_syns_per_cell
            log_all(
                logging.DEBUG,
                "%s: Average syns/cell: %.1f, Estimated total: %d ",
                metype,
                average_syns_per_cell,
                metype_estimate,
            )
            total_estimate += metype_estimate

        return int(total_estimate)

    def get_target_connections(
        self, src_target_name, dst_target_name, selected_gids=None, conn_population=None
    ):
        """Retrives the connections between src-dst cell targets

        Args:
             selected_gids: (optional) post gids to select (original, w/o offsetting)
             conn_population: restrict the set of connections to be returned
        """
        src_target_spec = TargetSpec(src_target_name)
        dst_target_spec = TargetSpec(dst_target_name)

        src_target = (
            self._target_manager.get_target(src_target_spec)
            if src_target_spec.name is not None
            else None
        )
        assert dst_target_spec.name, "No target specified for `get_target_connections`"
        dst_target = self._target_manager.get_target(dst_target_spec)
        if (src_target and src_target.is_void()) or dst_target.is_void():
            return

        tgid_offset = self.target_pop_offset
        conn_populations: list[ConnectionSet] = (
            (conn_population,) if conn_population is not None else self._populations.values()
        )

        # temporary set for faster lookup
        src_gids = src_target and set(src_target.gids(raw_gids=False))

        for population in conn_populations:
            logging.debug("Connections from population %s", population)
            tgids = np.fromiter(population.target_gids(), "uint32")
            tgids = np.intersect1d(tgids, dst_target.gids(raw_gids=False))
            if selected_gids:
                tgids = np.intersect1d(tgids, selected_gids + tgid_offset)
            for conn in population.get_connections(tgids):
                if src_target is None or conn.sgid in src_gids:
                    yield conn

    def configure_group(self, conn_config, gidvec=None):
        """Configure connections according to a config Connection block

        Args:
            conn_config: The connection configuration dict
            gidvec: A restricted set of gids to configure (original, w/o offsetting)
        """
        src_target = conn_config["Source"]
        dst_target = conn_config["Destination"]
        syn_params = dict_filter_map(
            conn_config,
            {
                "Weight": "weight_factor",
                "SpontMinis": "minis_spont_rate",
                "SynDelayOverride": "syndelay_override",
                "NeuromodStrength": "neuromod_strength",
                "NeuromodDtc": "neuromod_dtc",
            },
        )

        # Load eventual mod override helper
        if "ModOverride" in conn_config:
            logging.info("   => Overriding mod: %s", conn_config["ModOverride"])
            override_helper = conn_config["ModOverride"] + "Helper"
            Nd.load_hoc(override_helper)
            assert hasattr(Nd.h, override_helper), (
                "ModOverride helper doesn't define hoc template: " + override_helper
            )

        configured_conns = 0
        for conn in self.get_target_connections(src_target, dst_target, gidvec):
            for key, val in syn_params.items():
                setattr(conn, key, val)
            if "ModOverride" in conn_config:
                conn.mod_override = conn_config["ModOverride"]
            if "SynapseConfigure" in conn_config:
                conn.add_synapse_configuration(conn_config["SynapseConfigure"])
            configured_conns += 1
        return configured_conns

    def restart_events(self):
        """After restore, restart the artificial events (replay and spont minis)"""
        for conn in self.all_connections():
            conn.restart_events()

    def _unlock_all_connections(self):
        """Unlock all, mainly when we load a new connectivity source"""
        for conn in self.all_connections():
            conn.locked = False

    def finalize(self, base_seed=0, *, _conn_type="synapses", **conn_params):
        """Instantiates the netcons and Synapses for all connections.

        Note: All weight scalars should have their final values.

        Args:
            base_seed: optional argument to adjust synapse RNGs (default=0)
            _conn_type: (Internal) A string repr of the connectivity type
            conn_params: Additional finalize parameters for the specific _finalize_conns
                E.g. replay_mode (Default: Auto-Detect) Use DISABLED to skip replay
                and COMPLETE to instantiate VecStims in all synapses

        """
        self._synapse_reader = None  # Destroy to release memory (all cached params)
        logging.info("Instantiating %s... Params: %s", _conn_type, conn_params)
        n_created_conns = 0

        for popid, pop in self._populations.items():
            attach_src = pop.src_pop_id == 0 or not pop.virtual_source  # real populations
            conn_params["attach_src_cell"] = attach_src
            logging.info(
                " * Connections among %s -> %s, attach src: %s",
                pop.src_pop_name or "(base)",
                pop.dst_pop_name or "(base)",
                attach_src,
            )

            for tgid, conns in ProgressBar.iter(pop.items(), name="Pop:" + str(popid)):
                n_created_conns += self._finalize_conns(tgid, conns, base_seed, **conn_params)

        all_ranks_total = MPI.allreduce(n_created_conns, MPI.SUM)
        logging.info(" => Created %d %s", all_ranks_total, _conn_type)
        return all_ranks_total

    def _finalize_conns(self, tgid, conns, base_seed, *, reverse=False, **kwargs):
        """Low-level handling of finalizing connections belonging to a target gid.
        By default it calls finalize on each cell.
        """
        # Note: *kwargs normally contains 'replay_mode' but may differ for other types
        metype = self._cell_manager.get_cell(tgid)
        n_created_conns = 0
        if reverse:
            conns = reversed(conns)
        for conn in conns:  # type: Connection
            syn_count = conn.finalize(metype, base_seed, **kwargs)
            logging.debug("Instantiated conn %s: %d synapses", conn, syn_count)
            n_created_conns += syn_count
        return n_created_conns

    def replay(self, *_, **_kw):
        logging.warning("Replay is not available in %s", self.__class__.__name__)


# ##############
# Helper methods
# ##############


def edge_node_pop_names(edge_file, edge_pop_name, src_pop_name=None, dst_pop_name=None) -> tuple:
    """Find the node populations names.

    Args:
        edge_file: The edge file to extract the population names from
        edge_pop_name: The name of the edge population
    Returns: tuple of the src-dst population names.
    """
    src_dst_pop_names = _edge_meta_get_node_populations(edge_file, edge_pop_name)
    if src_dst_pop_names:
        if src_pop_name is None:
            src_pop_name = src_dst_pop_names[0]
        if dst_pop_name is None:
            dst_pop_name = src_dst_pop_names[1]
    return src_pop_name, dst_pop_name


@run_only_rank0
def _edge_meta_get_node_populations(edge_file, edge_pop_name) -> tuple:
    import libsonata

    edge_storage = libsonata.EdgeStorage(edge_file)
    if not edge_pop_name:
        assert len(edge_storage.population_names) == 1, (
            "multi-population edges require manual selection"
        )
        edge_pop_name = next(iter(edge_storage.population_names))

    edge_pop = edge_storage.open_population(edge_pop_name)
    return (edge_pop.source, edge_pop.target)


# ######################################################################
# SynapseRuleManager
# ######################################################################
class SynapseRuleManager(ConnectionManagerBase):
    """The SynapseRuleManager is designed to encapsulate the creation of
    synapses for BlueBrain simulations, handling the data coming from
    the circuit file. If the config file provides any Connection
    Rules, those override which synapses are created.

    Note that the Connection rules are processed with the assumption
    that they come in the config file from more general to more specific.
    E.g.: A column->column connection should come before
    layer 4 -> layer 2 which should come before L4PC -> L2PC.

    Once all synapses are prepared with final weights, the Netcons can be
    created.
    """

    CONNECTIONS_TYPE = ConnectionTypes.Synaptic

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        """Initializes a Connection/Edge manager for standard METype synapses

        Args:
            circuit_conf: The configuration object/dict
            target_manager: The (hoc-level) target manager
            cell_manager: The destination cell population manager
            src_cell_manager: The source cell population manager [Default: same as destination]
        """
        super().__init__(circuit_conf, target_manager, cell_manager, src_cell_manager, **kw)
        # SynapseRuleManager opens synapse file and init populations
        syn_source = circuit_conf.get("nrnPath")
        if syn_source:
            logging.info("Init %s. Options: %s", type(self).__name__, kw)
            self.open_edge_location(syn_source, circuit_conf, **kw)

    def finalize(self, base_seed=0, **kwargs):
        """Create the actual synapses and netcons. See super() docstring"""
        kwargs.setdefault("replay_mode", ReplayMode.AS_REQUIRED)
        super().finalize(base_seed, **kwargs)

    def _finalize_conns(self, tgid, conns, base_seed, **kw):
        # Note: (Compat) neurodamus hoc finalizes connections in reversed order.
        return super()._finalize_conns(tgid, conns, base_seed, reverse=True, **kw)

    def setup_delayed_connection(self, conn_config):
        """Setup delayed connection weights for synapse initialization.

        Find source and target gids and the associated connection,
        and add the delay and weight to their delay vectors.

        Args:
            conn_config: Connection configuration parsed from sonata config
        """
        src_target_name = conn_config["Source"]
        dst_target_name = conn_config["Destination"]
        delay = conn_config["Delay"]
        new_weight = conn_config.get("Weight", 0.0)

        configured_conns = 0
        for conn in self.get_target_connections(src_target_name, dst_target_name):
            conn.add_delayed_weight(delay, new_weight)
            configured_conns += 1
        return configured_conns

    @timeit(name="Replay inject")
    def replay(self, spike_manager, src_target_name, dst_target_name, start_delay=0.0):
        """Create special netcons to trigger timed spikes on those synapses.

        Args:
            spike_manager: map of gids (pre-synaptic) with spike times
            src_target_name: Source population:target of the replay connections
            dst_target_name: Target whose gids should be replayed
            start_delay: Dont deliver events before t=start_delay
        """
        log_verbose("Applying replay map with %d src cells...", len(spike_manager))
        replayed_count = 0

        # Dont deliver events in the past
        if Nd.t > start_delay:
            start_delay = Nd.t
            log_verbose("Restore: Delivering events only after t=%.4f", start_delay)

        src_pop_offset = self.src_pop_offset

        for conn in self.get_target_connections(src_target_name, dst_target_name):
            raw_sgid = conn.sgid - src_pop_offset
            if raw_sgid not in spike_manager:
                continue
            conn.replay(spike_manager[raw_sgid], start_delay)
            replayed_count += 1

        total_replays = MPI.allreduce(replayed_count, MPI.SUM)
        if MPI.rank == 0:
            if total_replays == 0:
                logging.warning("No connections were injected replay stimulus")
            else:
                logging.info(" => Replaying on %d connections", total_replays)
        return total_replays



from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from .utils.pyutils import StrEnumBase, cache_errors

if TYPE_CHECKING:
    from neurodamus.target_manager import TargetPointList


class ReportSetupError(Exception):
    pass


class SectionType(StrEnumBase):
    ALL = 0
    SOMA = 1
    AXON = 2
    DEND = 3
    APIC = 4
    INVALID = 5

    __mapping__ = [
        ("all", ALL),
        ("soma", SOMA),
        ("axon", AXON),
        ("dend", DEND),
        ("apic", APIC),
        ("invalid", INVALID),
    ]
    __default__ = SOMA
    __invalid__ = INVALID


class CompartmentType(StrEnumBase):
    ALL = 0
    CENTER = 1
    INVALID = 2

    __mapping__ = [
        ("all", ALL),
        ("center", CENTER),
        ("invalid", INVALID),
    ]
    __default__ = CENTER
    __invalid__ = INVALID


class Scaling(StrEnumBase):
    NONE = 0
    AREA = 1

    __mapping__ = [
        ("none", NONE),
        ("area", AREA),
    ]
    __default__ = AREA


class ReportType(StrEnumBase):
    COMPARTMENT = 0
    COMPARTMENT_SET = 1
    SUMMATION = 2
    SYNAPSE = 3
    LFP = 4

    __mapping__ = [
        ("compartment", COMPARTMENT),
        ("compartment_set", COMPARTMENT_SET),
        ("summation", SUMMATION),
        ("synapse", SYNAPSE),
        ("lfp", LFP),
    ]


@dataclass
class ReportParameters:
    type: ReportType
    name: str
    report_on: str
    unit: str
    format: str
    dt: float
    start: float
    end: float
    output_dir: str
    buffer_size: int
    scaling: Scaling
    target: object
    sections: SectionType
    compartments: CompartmentType
    compartment_set: str
    points: list[TargetPointList] | None = None  # this is filled later with get_point_list


@cache_errors
def check_report_parameters(
    rep_params: ReportParameters, nd_dt: float, *, lfp_active: bool
) -> None:
    """Validate report parameters against simulation constraints."""
    errors = []
    if rep_params.start > rep_params.end:
        errors.append(
            f"Invalid report configuration: end time ({rep_params.end}) is "
            f"before start time ({rep_params.start})."
        )

    if rep_params.dt < nd_dt:
        errors.append(
            f"Invalid report configuration: report dt ({rep_params.dt}) is smaller "
            f"than simulation dt ({nd_dt})."
        )

    if rep_params.type == ReportType.LFP and not lfp_active:
        errors.append(
            "LFP report setup failed: electrodes file may be missing or "
            "simulator is not set to CoreNEURON."
        )

    if errors:
        raise ReportSetupError("\n".join(errors))


@cache_errors
def create_report_parameters(sim_end, nd_t, output_root, rep_name, rep_conf, target, buffer_size):
    """Create report parameters from configuration."""
    start_time = rep_conf["StartTime"]
    end_time = rep_conf.get("EndTime", sim_end)
    rep_dt = rep_conf["Dt"]
    rep_type = ReportType.from_string(rep_conf["Type"])
    if nd_t > 0:
        start_time += nd_t
        end_time += nd_t
    end_time = min(end_time, sim_end)

    sections = SectionType.from_string(rep_conf.get("Sections"))
    compartments = CompartmentType.from_string(rep_conf.get("Compartments"))

    logging.info(
        " * %s (Type: %s, Target: %s, Dt: %f)",
        rep_name,
        rep_type,
        rep_conf["Target"],
        rep_dt,
    )

    return ReportParameters(
        type=rep_type,
        name=Path(rep_conf.get("FileName", rep_name)).name,
        report_on=rep_conf["ReportOn"],
        unit=rep_conf["Unit"],
        format=rep_conf["Format"],
        dt=rep_dt,
        start=start_time,
        end=end_time,
        output_dir=output_root,
        buffer_size=buffer_size,
        scaling=Scaling.from_string(rep_conf.get("Scaling")),
        target=target,
        sections=sections,
        compartments=compartments,
        compartment_set=rep_conf.get("CompartmentSet"),
    )



import logging

from .core import NeuronWrapper as Nd
from .report_parameters import ReportParameters, ReportType, Scaling, SectionType
from .utils.pyutils import cache_errors


class Report:
    """Abstract base class for handling simulation reports in NEURON.

    Provides methods for parsing report parameters, handling variables, and defining the structure
    required by subclasses to append specific data (e.g., compartments or currents).
    """

    CURRENT_INJECTING_PROCESSES = {"SEClamp", "IClamp"}

    def __init__(
        self,
        params,
        use_coreneuron,
    ):
        """Initialize a Report object with simulation parameters.

        :param params: Object containing report configuration (e.g., name, dt, unit).
        :param use_coreneuron: Boolean indicating if CoreNEURON is enabled.
        """
        if type(self) is Report:
            raise TypeError("Report is an abstract base class and cannot be instantiated directly.")

        self.type = params.type
        self.variables = self.parse_variable_names(params.report_on)

        self.report_dt = params.dt
        self.scaling = params.scaling
        self.use_coreneuron = use_coreneuron

        self.alu_list = []
        self.report = Nd.SonataReport(
            0.5,
            params.name,
            params.output_dir,
            params.start,
            params.end,
            params.dt,
            params.unit,
            "compartment",
        )
        Nd.BBSaveState().ignore(self.report)

    def register_gid_section(
        self, cell_obj, point, vgid, pop_name, pop_offset, sections: SectionType
    ):
        """Abstract method to be implemented by subclasses to add section-level report data.

        :raises NotImplementedError: Always, unless overridden in subclass.
        """
        raise NotImplementedError("Subclasses must implement register_gid_section()")

    @cache_errors
    def setup(self, rep_params: ReportParameters, global_manager):
        for point in rep_params.points:
            gid = point.gid
            pop_name, pop_offset = global_manager.getPopulationInfo(gid)
            cell = global_manager.get_cell(gid)
            spgid = global_manager.getSpGid(gid)

            self.register_gid_section(cell, point, spgid, pop_name, pop_offset, rep_params.sections)

    @staticmethod
    def is_point_process_at_location(point_process, section, x):
        """Check if a point process is located at a specific position within a section.

        :param point_process: The point process to check.
        :param section: The NEURON section in which the point process is located.
        :param x: The normalized position (0 to 1) within the section to check against.
        :return: True if the point process is at the specified position, False otherwise.
        """
        # Get the location of the point process within the section
        # warning: this pushes sec into neuron stack. Remember to pop_section!
        dist = point_process.get_loc()
        # Pop immediately after get_loc to avoid requiring caller to remember
        Nd.pop_section()
        # Calculate the compartment ID based on the location and number of segments
        compartment_id = int(dist * section.nseg)
        # Check if the compartment ID matches the desired location
        return compartment_id == int(x * section.nseg)

    @staticmethod
    def get_point_processes(section, mechanism=None):
        """Retrieve all synapse objects attached to a given section.

        :param section: The NEURON section object to search for synapses.
        :param mechanism: The mechanism requested
        :return: A list of synapse objects attached to the section.
        """
        if mechanism is None:
            return [syn for seg in section for syn in seg.point_processes()]

        return [
            syn
            for seg in section
            for syn in seg.point_processes()
            if syn.hname().startswith(mechanism)
        ]

    @staticmethod
    def parse_variable_names(report_on):
        """Parse variable names from a user-specified string into mechanism-variable tuples.

        Also, activate fast_i_mem if necessary.

        E.g., "hh.ina pas.i" → [("hh", "ina"), ("pas", "i")]

        :return: List of (mechanism, variable) tuples.
        """
        tokens_with_vars = []
        tokens = report_on.split()  # Splitting by whitespace

        for val in tokens:
            if "." in val:
                mechanism, var = val.split(".", 1)  # Splitting by the first period
                tokens_with_vars.append((mechanism, var))
            else:
                if val == "i_membrane":
                    Nd.cvode.use_fast_imem(1)
                    val = "i_membrane_"
                tokens_with_vars.append((val, "i"))  # Default internal variable

        return tokens_with_vars

    @staticmethod
    def get_var_refs(section, x, mechanism, variable_name):
        """Retrieve references to a variable within a mechanism at a specific location on a section.

        This method returns a list of variable references (_ref_<variable_name>) either from:
        - The point processes of the given mechanism located at position x on the section, or
        - Directly from the section or its inserted mechanism if no point processes are present.

        Parameters:
        section : h.Section
            The NEURON section to query.
        x : float
            The location along the section (0 <= x <= 1).
        mechanism : str
            The name of the mechanism or point process.
        variable_name : str
            The name of the variable whose reference is requested.

        Returns:
        list
            A list of variable references (typically hoc references) matching the query.
        """
        point_processes = Report.get_point_processes(section, mechanism)
        # if not a point process, it is a current of voltage. Directly return the reference

        if not point_processes:
            sec_x = section(x)
            var_name = "_ref_" + mechanism
            if hasattr(sec_x, var_name):
                return [getattr(sec_x, var_name)]
            if hasattr(sec_x, mechanism):
                mech = getattr(sec_x, mechanism)
                var_name = "_ref_" + variable_name
                if hasattr(mech, var_name):
                    return [getattr(mech, var_name)]
            return []
        # search among the point processes the ones that at at position x and return the reference
        return [
            getattr(pp, "_ref_" + variable_name)
            for pp in point_processes
            if Report.is_point_process_at_location(pp, section, x)
            and hasattr(pp, "_ref_" + variable_name)
        ]

    def get_scaling_factor(self, section, x, mechanism):
        """Scaling factors for some special variables"""
        if mechanism in self.CURRENT_INJECTING_PROCESSES:
            return -1.0  # Negative for current injecting processes

        if mechanism != "i_membrane_" and self.scaling == Scaling.AREA:
            return section(x).area() / 100.0

        return 1.0


class CompartmentReport(Report):
    """Concrete Report subclass for reporting compartment-level variables.

    Appends variable references at specific compartment locations for a given cell.
    """

    def __init__(
        self,
        params,
        use_coreneuron,
    ):
        super().__init__(params=params, use_coreneuron=use_coreneuron)
        if len(self.variables) != 1:
            raise ValueError(
                f"Compartment reports requires exactly one variable, "
                f"but received: `{self.variables}`"
            )

    def register_gid_section(
        self, cell_obj, point, vgid, pop_name, pop_offset, _sections: SectionType
    ):
        """Append section-based report data for a single cell and its compartments.

        :param cell_obj: The cell being processed.
        :param point: Point data containing section list and location.
        :param vgid: Virtual GID to use in report.
        :param pop_name: Population name.
        :param pop_offset: Offset for population indexing.
        :param _sum_currents_into_soma: Unused parameter in this subclass.
        """
        if self.use_coreneuron:
            return
        gid = cell_obj.gid
        vgid = vgid or gid

        mechanism, variable_name = self.variables[0]
        self.report.AddNode(gid, pop_name, pop_offset)
        for i, sc in enumerate(point.sclst):
            section = sc.sec
            x = point.x[i]
            var_refs = self.get_var_refs(section, x, mechanism, variable_name)
            if len(var_refs) == 0:
                raise AttributeError(
                    f"No reference found for variable '{variable_name}' of mechanism '{mechanism}' "
                    f"at location {x}."
                )

            if len(var_refs) > 1:
                raise AttributeError(
                    f"Expected one reference for variable '{variable_name}' "
                    f"of mechanism '{mechanism}' at location "
                    f"{x}, but found {len(var_refs)}. "
                    "Probably many synapses attached to the soma. "
                    "Compartment reports require only one variable per segment."
                )
            section_id = cell_obj.get_section_id(section)
            self.report.AddVar(var_refs[0], section_id, gid, pop_name)


class SummationReport(Report):
    """Concrete Report subclass for summing currents or other variables across sections.

    Handles intrinsic currents and point processes, possibly summing them into soma.
    """

    def register_gid_section(
        self, cell_obj, point, vgid, pop_name, pop_offset, sections: SectionType
    ):
        """Append summed variable data for a given cell across sections.

        :param cell_obj: The cell being reported.
        :param point: Point containing section list and x positions.
        :param vgid: Optional virtual GID.
        :param pop_name: Population name.
        :param pop_offset: Population GID offset.
        :param sections: Sum into soma if section is soma

        Note: sections == SectionType.SOMA effectively means that we need
        to sum the values into the soma
        """
        if self.use_coreneuron:
            return
        gid = cell_obj.gid
        vgid = vgid or gid

        self.report.AddNode(gid, pop_name, pop_offset)

        if sections == SectionType.SOMA:
            alu_helper = self.setup_alu_for_summation(0.5)

        for i, sc in enumerate(point.sclst):
            section = sc.sec
            x = point.x[i]
            if sections == SectionType.ALL:
                alu_helper = self.setup_alu_for_summation(x)

            self.process_mechanisms(section, x, alu_helper)

            if sections == SectionType.ALL:
                section_index = cell_obj.get_section_id(section)
                self.add_summation_var_and_commit_alu(alu_helper, section_index, gid, pop_name)
        if sections == SectionType.SOMA:
            # soma
            self.add_summation_var_and_commit_alu(alu_helper, 0, gid, pop_name)

    def process_mechanisms(self, section, x, alu_helper):
        """Add the ref variable identified by x, mechanism, and variable_name
        to alu_helper multiplied by the scaling_factor.

        Note: compartments without the variable are silently skipped.
        """
        for mechanism, variable in self.variables:
            scaling_factor = self.get_scaling_factor(section, x, mechanism)
            if scaling_factor == 0:
                logging.warning(
                    "Skipping intrinsic current '%s' at a location with area = 0", mechanism
                )
                # simply skip if not valid. No logging or error is necessary
                continue
            var_refs = Report.get_var_refs(section, x, mechanism, variable)
            for var_ref in var_refs:
                alu_helper.addvar(var_ref, scaling_factor)

    def setup_alu_for_summation(self, alu_x):
        """Setup ALU helper for summation."""
        alu_helper = Nd.ALU(alu_x, self.report_dt)
        alu_helper.setop("summation")
        bbss = Nd.BBSaveState()
        bbss.ignore(alu_helper)
        return alu_helper

    def add_summation_var_and_commit_alu(self, alu_helper, section_index, gid, population_name):
        """Add the ALU's output as a summation variable and commit it to the report."""
        self.report.AddVar(alu_helper._ref_output, section_index, gid, population_name)
        # Append ALUhelper to the list of ALU objects
        self.alu_list.append(alu_helper)


class SynapseReport(Report):
    def __init__(
        self,
        params,
        use_coreneuron,
    ):
        super().__init__(params=params, use_coreneuron=use_coreneuron)
        if len(self.variables) != 1:
            raise ValueError(
                f"Synapse reports requires exactly one variable, but received: `{self.variables}`"
            )

    def register_gid_section(
        self, cell_obj, point, vgid, pop_name, pop_offset, sections: SectionType
    ):
        """Append synapse variables for a given cell to the report grouped by gid."""
        gid = cell_obj.gid
        # Default to cell's gid if vgid is not provided
        vgid = vgid or cell_obj.gid

        # Initialize lists for storing synapses and their locations
        synapse_list = []
        mechanism, variable = self.variables[0]
        # Evaluate which synapses to report on
        for i, sc in enumerate(point.sclst):
            section = sc.sec
            x = point.x[i]
            # Iterate over point processes in the section

            point_processes = self.get_point_processes(section, mechanism)
            for synapse in point_processes:
                if self.is_point_process_at_location(synapse, section, x):
                    synapse_list.append(synapse)
                    # Mark synapse as selected for report
                    if hasattr(synapse, "selected_for_report"):
                        synapse.selected_for_report = True

        if not synapse_list:
            raise AttributeError(f"Synapse '{mechanism}' not found on any points on gid: {gid}. ")
        if not self.use_coreneuron:
            # Prepare the report for the cell
            self.report.AddNode(gid, pop_name, pop_offset)
            try:
                for synapse in synapse_list:
                    var_ref = getattr(synapse, "_ref_" + variable)
                    self.report.AddVar(var_ref, synapse.synapseID, gid, pop_name)
            except AttributeError as e:
                msg = f"Variable '{variable}' not found at '{synapse.hname()}'."
                raise AttributeError(msg) from e


NOT_SUPPORTED = object()
_report_classes = {
    ReportType.COMPARTMENT: CompartmentReport,
    ReportType.COMPARTMENT_SET: CompartmentReport,
    ReportType.SUMMATION: SummationReport,
    ReportType.SYNAPSE: SynapseReport,
    ReportType.LFP: NOT_SUPPORTED,
}


@cache_errors
def create_report(params: ReportParameters, use_coreneuron):
    """Factory function to create a report instance based on parameters."""
    cls = _report_classes.get(params.type)
    if cls is None:
        raise ValueError(f"Unknown report type: {params.type.to_string()}")
    if cls is NOT_SUPPORTED:
        return None
    return cls(params, use_coreneuron)
